from ..io.yaml_functions import load_yaml
import yaml
import os


def compile_modality_parameters(
    modality_id: list, congifuration_directory, fluo_emissions=None, prints=False
):
    """
    args:
        modality_id: (string) ID of modality
        congifuration_directory: (string) Path of configuration settings
        emission: (string) Emission type, either constant, or blinking.

    returns:
        modality_optical_pars: dictionary of parameters per modality
    """
    # load modality configuration file
    mode_file = modality_id + ".yaml"
    modality_config_path = os.path.join(
        congifuration_directory, "modalities", mode_file
    )
    mod_pars = load_yaml(modality_config_path)
    # approximate psf stack size
    psf_params = dict(
        stack_source=mod_pars["PSF"]["source"],
        scale=mod_pars["scale"],
        shape=[150, 150, 150],  # arbitrary, ideally should cover the shole PSF
        std_devs=[
            mod_pars["PSF"]["resolution"]["X"] / mod_pars["PSF"]["voxelsize"],
            mod_pars["PSF"]["resolution"]["Y"] / mod_pars["PSF"]["voxelsize"],
            mod_pars["PSF"]["resolution"]["Z"] / mod_pars["PSF"]["voxelsize"],
        ],
        voxelsize=[
            mod_pars["PSF"]["voxelsize"],
            mod_pars["PSF"]["voxelsize"],
            mod_pars["PSF"]["voxelsize"],
        ],
        depth=int(mod_pars["depth"] / mod_pars["PSF"]["voxelsize"]),
    )
    factor_ = 1000  # to match detector params to imager, needs fix
    detector_params = dict(
        scale=1.0e-06,
        pixelsize=mod_pars["detector"]["pixelsize"] / factor_,
        noise_model=dict(
            binomial={"p": mod_pars["detector"]["noise"]["binomial"]},
            gamma={"g": mod_pars["detector"]["noise"]["gain"]},
            baselevel={"bl": mod_pars["detector"]["noise"]["baselevel"]["mean"]},
            gaussian={
                "sigma": mod_pars["detector"]["noise"]["gaussian"]["standard_dev"]
            },
            conversion={"adu": mod_pars["detector"]["noise"]["ADU"]},
        ),
        noise_order=["binomial", "gamma", "baselevel", "gaussian", "conversion"],
        bits_pixel=32,
    )
    # get filters default should be one per channel
    if fluo_emissions is None:
        filter_dictionary = None
        emission_behaviour = None
        modality_name = mod_pars["name"]
        modality_params = dict(
            filters=filter_dictionary,
            psf_params=psf_params,
            detector=detector_params,
            emission=emission_behaviour,  # blinking or constant
            modality=modality_name,
        )
        return modality_params, mod_pars
    else:
        if (
            "filters" not in mod_pars.keys()
        ):  # each fluorophore will be assign and independent channel
            if prints:
                print("Creating channel for each fluorophore")
            ch = 0
            filter_dictionary = dict()
            for fluoname in list(fluo_emissions.keys()):
                if prints:
                    print(fluoname)
                channel_name = "ch" + str(ch)
                fluorophores_in_channel = []
                fluorophores_in_channel.append(
                    fluoname
                )  # this is because a channel can have several fluorophores
                filter_dictionary[channel_name] = fluorophores_in_channel
                ch += 1
        else:
            filter_dictionary = mod_pars["filters"]
        # get emission type from fluorophore
        emission_behaviour = define_emission_behaviour("constant", fluo_emissions)
        print(f"emission for {str(modality_id)} set to {emission_behaviour}")
        # assign modality name
        modality_name = mod_pars["name"]
        modality_params = dict(
            filters=filter_dictionary,
            psf_params=psf_params,
            detector=detector_params,
            emission=emission_behaviour,  # blinking or constant
            modality=modality_name,
        )
        return modality_params, mod_pars


def define_emission_behaviour(mod_emission, fluo_emissions):
    """
    Function that checks if the modality emission and fluorophores emission match
    """
    emission = None
    if mod_emission == "constant":
        emission = "constant"
    elif mod_emission == "blinking":
        emission = "blinking"
        for key, value in fluo_emissions.items():
            if value == "constant":
                emission = "constant"

    return emission


def format_modality_acquisition_params(
    exp_time=0.001,
    noise=True,
    save=False,
    nframes=10,
    channels=[
        "ch0",
    ],
    **kwargs,
):
    """
    Create expected format for acquisition parameters
    """
    mod_acquisition_params = dict(
        exp_time=exp_time, noise=noise, save=save, nframes=nframes, channels=channels
    )
    return mod_acquisition_params


def write_parameters(
    structure=None,
    probes=None,
    virtualsample=None,
    modalities=None,
    acquisition=None,
    writingpath: str = None,
    **kwargs,
):
    if writingpath:
        virtualsample.pop("field_emitters")
        virtualsample.pop("reference_point")
        saving_path = os.path.join(writingpath, "parameters.yml")
        with open(saving_path, "w") as outfile:
            yaml.dump_all(
                [structure, probes, virtualsample, modalities, acquisition],
                outfile,
                default_flow_style=False,
            )



import numpy as np
import copy
from scipy.spatial.transform import Rotation as R  

# # Rodrigues Rotation
def rotate_point(point, axis, angle):
    ## implements the Rodrigues' rotation formula
    cos_theta = np.cos(angle)
    sin_theta = np.sin(angle)
    axis = axis / np.linalg.norm(axis)
    cross = np.cross(axis, point)
    dot = np.dot(axis, point)
    return cos_theta * point + sin_theta * cross + (1 - cos_theta) * dot * axis


def rotate_set(points, axi, angl):
    rotated_pts = []
    for pt in range(len(points)):
        rotated_pts.append(rotate_point(points[pt], axi, angl))
    rotated_pts = np.array(rotated_pts)
    return rotated_pts


# following function was reorient_set
def labeling_reorient_set(label_points, piv_id, intern_axis_id, direction, end_point):
    # pts are the points to reorient
    # piv_id is the pivot index that will be use to reorient pts
    # intern_axis_id is an index for the other point
    # in pts that will define its internal axix of rotation
    # this also dictates the orientation of the internal axis so that
    # direction sets the new direction all the points need to
    # follow relative to its internal axis
    # ref is the reference point for overall traslation of pts
    # Translate towards the origin
    # translated_origin, displacement = displace_set2(pts,piv_id,[0,0,0])
    translated_origin, displacement = transform_displace_set(
        label_points, label_points[piv_id], np.array([0, 0, 0])
    )
    # define an internal axis of orientation
    int_axis = translated_origin[intern_axis_id] - translated_origin[piv_id]
    # get rotation axis
    V = np.cross(int_axis, direction)
    if np.linalg.norm(V) == 0:
        V = direction
    # get angle or rotation in radians
    theta = np.arccos(
        np.dot(int_axis, direction)
        / (np.linalg.norm(int_axis) * np.linalg.norm(direction))
    )
    # Rotate pts
    rotated_at_origin = rotate_set(translated_origin, V, theta)
    reoriented, _ = transform_displace_set(
        rotated_at_origin, rotated_at_origin[piv_id], end_point
    )
    return reoriented, V


def transform_displace_set(
    points, external_reference, end_point
):  # previously displace_set_3
    # points are the set of points to displace
    # external_reference is the reference point
    # that usually is inside the cloud of points
    ## end_point is where the external_reference will be relocated, so that
    # a translation vector for all points is defined
    transl_vec = end_point - external_reference  # specify the translation vector
    translated = points + transl_vec
    return translated, transl_vec


def rotate_pts_by_vector(pts, current_vector, new_vector, pts_referece_center):
    translated_origin, displacement = transform_displace_set(
        pts, pts_referece_center, np.array([0, 0, 0])
    )
    # get rotation axis by calculating the cross product of the current axis with new
    ax_of_rot = np.cross(current_vector, new_vector)
    if np.linalg.norm(ax_of_rot) == 0:
        ax_of_rot = new_vector
    # get angle or rotation in radians
    theta = np.arccos(
        np.dot(current_vector, new_vector)
        / (np.linalg.norm(current_vector) * np.linalg.norm(new_vector))
    )
    # Rotate pts
    rotated = rotate_set(translated_origin, ax_of_rot, theta)
    reoriented, _ = transform_displace_set(
        rotated, np.array([0, 0, 0]), pts_referece_center
    )
    return reoriented


# method that takes the epitopes with their corresponding normals
# and takes the labeling entity generated with gen_labeling_entity
def decorate_epitopes_normals(normals_ft_epitopes, labeling_entity, dol:int = None, **kwargs):
    """
    labeling_entity is assumed to have the pivot and the axis orientation
    as the first two points, only after that are defined the actual emitters
    they will be removed after decorating the epitopes

    output: the new positioned emitters without pivots
    """
    labeling_entity_copy = copy.copy(labeling_entity)
    pivot_reference_index = 0
    axis_defining_index = 1
    list_reoriented_points = []
    list_reoriented_points_normals = []
    for repl in range(np.shape(normals_ft_epitopes)[1]):
        new_points, rot_axis = labeling_reorient_set(
            labeling_entity_copy,
            pivot_reference_index,
            axis_defining_index,
            normals_ft_epitopes[0][repl],
            normals_ft_epitopes[1][repl],
        )
        # model degree of labelling
        if dol is not None:
            int_dol = np.random.poisson(lam=dol)
            print(f"DOL = {int_dol}")
            max_emitters = len(new_points)
            if int_dol != 0:
                if int_dol > max_emitters - 2 :
                    # use max value
                    list_reoriented_points.append(new_points)
                else:
                    # take the first two points to keep pivot and axis
                    new_points_dol = copy.copy(new_points[0:2])
                    # then randomly select the rest of the emitters
                    emitter_indices = np.arange(2, max_emitters)
                    selected_emitters = np.random.choice(
                        emitter_indices, size=int_dol, replace=False
                    )
                    for se in selected_emitters:
                        new_points_dol = np.vstack((new_points_dol, new_points[se]))
                    list_reoriented_points.append(new_points_dol)
                    #new_points = new_points[0:(2+int_dol)]
                    #list_reoriented_points.append(new_points)
        else:
            # print(new_points, rot_axis)
            list_reoriented_points.append(new_points)
            list_reoriented_points_normals.append(normals_ft_epitopes[0][repl])
    #print(f"before cleaning: {len(list_reoriented_points)}")
    # cleanup label entities and leave only the true emitters
    cleaned_up_labels = cleanup_labeling_entities(list_reoriented_points)
    return cleaned_up_labels, list_reoriented_points_normals


def cleanup_labeling_entities(labels_on_epitopes):
    """
    after repositioning and reotienting labels the list of labels
    is assumed to contain pivot and axis point on positions 0 and 1
    THis method removes those leaving only the true emitters
    """
    n_epitopes = len(labels_on_epitopes)
    # print(f"cleaning {n_epitopes}")
    pivot_reference_index = 0
    # prepare the list with the first entity's emitters
    fluorophores_xyz = labels_on_epitopes[pivot_reference_index][
        2:
    ]  # the first two are to be removed
    for i in range(n_epitopes - 1):
        # joining them like this allows for different number of emitters per antibody
        fluorophores_xyz = np.vstack(
            (fluorophores_xyz, labels_on_epitopes[i + 1][2:])
        )  # from i+1 since we already added the first
    return fluorophores_xyz


def planar_rotation(points, points_reference, unitary_vector_of_rotation, theta):
    """ 
    Rotate a set of points around a given rotation axis.

    Parameters
    ----------
    points : Nx3 array
        Array containing the N points to rotate
    points_reference : 1x3 array
        reference point for the set 
    unitary_vector_of_rotation: 1x3 array
        Vector at the origin to define the rotation axis
    theta: float
        Angle for rotation, in radians

    Returns
    -------
    numpy.ndarray or None
        Emitter coordinates or None if not found.
    """
    translated_origin, displacement = transform_displace_set(
        points, points_reference, np.array([0, 0, 0])
    )
    rotated_at_origin = rotate_set(translated_origin, unitary_vector_of_rotation, theta)
    rotated, _ = transform_displace_set(
        rotated_at_origin, np.array([0, 0, 0]), points_reference
    )
    return rotated
  
def apply_euler_rotation(vector, phi=0, theta=0, psi=0, order = "zyx", reset_orientation=False, **kwargs):
    """
    Rotations are extrinsic.

    
    theta: rotation around x axis in degrees
    phi: rotation around z axis in degrees
    psi: rotation around y axis in degrees
    """
    combined_R = R.from_euler(order, [phi,theta, psi], degrees=True)
    new_vector = combined_R.apply(vector)
    return new_vector


import numpy as np


def sample_spherical_normalised(npoints, ndim=3):
    vec = np.random.randn(ndim, npoints)
    vec /= np.linalg.norm(vec, axis=0)
    return vec.reshape(-1)


def boolean_epitope_selection(epitopes, selected_epitopes, new_epitope, min_distance):
    # epitopes is the numpy array containing epitopes coordinates
    # selected epitopes is a list of indices of the selected epitopes
    # min_distance is the minimum distance between the epitopes
    # function should return indices of the selected epitopes
    n_epi = len(selected_epitopes)
    for i in range(n_epi):
        is_available = 1
        if (
            np.linalg.norm(epitopes[selected_epitopes[i]] - epitopes[new_epitope])
            < min_distance
        ):
            is_available = 0
            break
    return is_available


def sample_epitopes_sterically(epitopes, min_distance):
    # epitopes is the coordinates array for the epitopes
    # min_distance is the minimum eculidean distance between the epitopes
    # step 1: select an epitope randomly
    selected_indices = []
    # randomize indices to have a guaranteed stop condition
    randomized_indices = np.random.choice(
        range(epitopes.shape[0]), epitopes.shape[0], replace=False
    )
    selected_indices.append(randomized_indices[0])
    for i in randomized_indices[1:]:
        # i here contains the index of the next epitope
        # verify is the next epitope is within the minimum distance
        if boolean_epitope_selection(epitopes, selected_indices, i, min_distance):
            selected_indices.append(i)
    # function should return indices of the selected epitopes
    return selected_indices


def sample_array(array, fraction=1):
    # assumes array has shape Nx3, so it will sample rows
    if fraction < 1:
        size = array.shape[0]
        ids = np.random.choice(np.arange(0, size), int(size * fraction), replace=False)
        subset = array[ids, :]
        return subset
    elif fraction == 1:
        return array


def binomial_epitope_sampling(epitopes, p=1, normals=None, min_distance=0.0):
    if min_distance:
        #print(f"sampling with min epitope distance: {min_distance}, and efficiency: {p}")
        available_epitopes_id = sample_epitopes_sterically(epitopes, min_distance)
        n_epitopes = len(available_epitopes_id)
    else:
        #print(f"sampling with efficiency: {p}")
        n_epitopes = epitopes.shape[0]
        available_epitopes_id = np.arange(0, n_epitopes)
    # bernoulli trials for each epitope 
    ids_selected = [x for x in available_epitopes_id if np.random.rand() < p]
    #ids_selected = np.random.binomial(available_epitopes_id, p)
    subset_epitopes = epitopes[ids_selected, :]
    n_epitopes_ = len(ids_selected)
    if normals is None:
        return subset_epitopes, n_epitopes_,  None
    else:
        subset_normals = normals[ids_selected, :]
        return subset_epitopes, n_epitopes_, subset_normals

def get_random_pixels(binary_image, num_pixels=3, min_distance=2):
    # Find indices of all positive pixels (value = 1)
    positive_pixels = np.argwhere(binary_image > 0)
    selected_pixels = []
    np.random.shuffle(positive_pixels)
    for pixel in positive_pixels:
        selected = True
        for i in range(len(selected_pixels)):
          if np.linalg.norm(pixel - selected_pixels[i]) < min_distance:
             selected = False
             break
        if selected:
            selected_pixels.append(pixel)
        if len(selected_pixels) >=  num_pixels:
           break
    return selected_pixels


import numpy as np
import copy
import matplotlib.pyplot as plt
import yaml
from ..analysis import metrics
import math

from .labelled_instance import LabeledInstance
from .molecular_structure import MolecularReplicates
from .labels import Label

from ..utils.transform.points_transforms import transform_displace_set
from ..utils.sample import arrays as sampl
from ..utils.data_format.visualisation import format_coordinates
from ..utils.visualisation.matplotlib_plots import add_ax_scatter


class Field:
    def __init__(self):
        """
        Initialize a Field object with default parameters for dimensions, scale, and molecule placement.
        """
        self.params = {}
        self.params["dimension_sizes"] = [1000.0, 1000.0, 10.0]
        self.params["scale"] = 1e-9
        self.params["relative_reference_point"] = [0.5, 0.5, 0.01]
        self.params["absolute_reference_point"] = None

        self.molecules = None

        self.molecules_params = {}
        self.molecules_params["nMolecules"] = 1
        self.molecules_params["relative_positions"] = np.array(
            [
                [0.5, 0.5, 0.01],
            ]
        ).reshape((1, 3))
        self.molecules_params["absolute_positions"] = None
        self.molecules_params["orientations"] = None
        self.molecules_params["orientations_planewise"] = None
        self.molecules_params["rotations"] = None
        self.molecules_params["minimal_distance"] = None
        self.xy_orientations = None
        self.xz_orientations = None
        self.yz_orientations = None
        self.axial_offset = None
        self.emitters_per_fluorophore = {}  # this can be exported

        #
        self.ximage_dimensions = None  # fov_info["imagesize"][0]  # in pixels
        self.yimage_dimensions = None  # fov_info["imagesize"][1]  # in pixels
        self.zimage_dimensions = None  # fov_info["imagesize"][2]  # in pixels
        self.z_offset = 0  # in scale units
        self.pixelsize = None  # fov_info["pixelsize"]  # in scale units
        self.relative_pos = None  # fov_info["relative_positions"]
        self.orientations = []
        self.absolute_pos = []
        self.fov_relative_ref_pt_xy = np.array([0.5, 0.5, 0.01])
        self.fov_absolute_ref_pt_xyself = None
        self.scale = 1e-9  # in meters
        self.particles = []
        self.label_names = []
        self.colours = {}
        self.fluorophre_emitters = {}
        self.fluo2labels = []
        self.plotting_params = dict()
        self.fluoparams = dict()
        self.random_placing = False
        self.random_orientations = False
        self.random_rotations = False
        self.sample_initial_orientation = None
        self.rotation_angles = None
        # print(f'Working scale of the Field of View is {self.scale} meters')

    # methods to initialise field parameteres
    def _set_dimension_sizes(self, dims):
        # limit x y and z in units of scale
        self.params["dimension_sizes"] = dims

    def _set_relative_reference(self, rel_ref: list):
        self.params["relative_reference_point"] = np.array(rel_ref)

    def _set_absolute_reference(self, abs_ref: np.ndarray):
        self.params["absolute_reference_point"] = abs_ref

    def _set_scale(self, scale: float):
        self.params["scale"] = scale

    def set_params(self, **kwargs):
        """
        Set field parameters using keyword arguments.

        Parameters
        ----------
        **kwargs
            Field parameter names and values.
        """
        for key, value in kwargs.items():
            self.params[key] = value

    def init_from_file(self, field_yaml):
        """
        Initialize field parameters from a YAML configuration file.

        Parameters
        ----------
        field_yaml : str
            Path to the YAML file containing field parameters.
        """
        with open(field_yaml, "r") as f:
            field_params = yaml.safe_load(f)
        params = dict(field_params["params"])
        molecules = dict(field_params["molecules_params"])
        export = dict(field_params["export_params"])
        # print("Field parameters for placing and orientation of labelled particles:")
        for key, value in molecules.items():
            print(key, ": ", value)
        # set general parameters
        self.set_params(**params)
        # set molecules params
        self.set_molecules_params(**molecules)

    def create_minimal_field(
        self, nmolecules=1, random_placing=False, random_orientations=False, random_rotations=False, **kwargs
    ):
        """
        Create a minimal field with a specified number of molecules and placement options.

        Parameters
        ----------
        nmolecules : int, optional
            Number of molecules to place in the field. Default is 1.
        random_placing : bool, optional
            If True, place molecules randomly. Default is False.
        random_orientations : bool, optional
            If True, assign random orientations to molecules. Default is False.
        **kwargs
            Additional parameters for field creation.
        """
        fluo_name = "AF647"
        self.calculate_absolute_reference()
        if "minimal_distance" in kwargs.keys():
            self.molecules_params["minimal_distance"] = kwargs["minimal_distance"]
        if "sample_dimensions" in kwargs.keys():
            self._set_dimension_sizes(kwargs["sample_dimensions"])
            self.calculate_absolute_reference()
        # positioning of particles in the field
        if "relative_positions" in kwargs.keys():
            self.random_placing = False
            print("Using relative positons to initialise field")
            nmolecules = len(kwargs["relative_positions"])
            self.set_molecule_param("nMolecules", nmolecules)
            self.molecules_params["relative_positions"] = kwargs["relative_positions"]
            self._gen_abs_from_rel_positions()
            self.fluorophre_emitters = {
                fluo_name: self.get_molecule_param("absolute_positions")
            }
            # ignore nmolecules and randomplacing
            # use thise values to set particles
        elif nmolecules > 1:
            print("More than 1 position needed. Randomising.")
            self.set_molecule_param("nMolecules", nmolecules)
            # if no list was passed, then we will use the default nmolecules parameter
            # having more than one particle necesarily use random placing
            self.random_placing = True
            self.generate_random_positions()
            self._gen_abs_from_rel_positions()
            self.fluorophre_emitters = {
                fluo_name: self.get_molecule_param("absolute_positions")
            }
        else:
            # if none of the above, this is the case of a single particle in the center
            # point = self.get_field_param("absolute_reference_point")
            self.set_molecule_param("nMolecules", 1)
            if random_placing:
                self.random_placing = True
                self.generate_random_positions()
            self._gen_abs_from_rel_positions()
            point = self.get_molecule_param("absolute_positions")
            self.fluorophre_emitters = {fluo_name: point.reshape(1, 3)}
        self._set_fluo_plotting_params(fluo_name)
        # Orientation of each particle 
        if random_orientations:
            self.random_orientations = True
            if "sample_inital_orientation" in kwargs.keys():
                self.sample_initial_orientation = kwargs["sample_inital_orientation"]
            if "xy_orientations" in kwargs.keys():
                self.xy_orientations = kwargs["xy_orientations"]
            if "xz_orientations" in kwargs.keys():
                self.xz_orientations = kwargs["xz_orientations"]
            if "yz_orientations" in kwargs.keys():
                self.yz_orientations = kwargs["yz_orientations"]
        else:
            self.random_orientations = False
            if "orientation_per_particle" in kwargs.keys():
                self.set_molecule_param("orientation_per_particle", kwargs["orientation_per_particle"])
        # rotation within plane of each particle 
        if random_rotations:
            self.random_rotations = True
            if "rotation_angles" in kwargs.keys():
                self.rotation_angles = kwargs["rotation_angles"]
        else:
            self.random_rotations = False
            if "rotation_per_particle" in kwargs.keys():
                self.set_molecule_param("rotation_per_particle", kwargs["rotation_per_particle"])
        # global axial offset 
        if "axial_offset" in kwargs.keys():
            self.axial_offset = kwargs["axial_offset"]


    def calculate_absolute_reference(self):
        """
        Calculate and set the absolute reference point based on the relative reference and dimension sizes.
        """
        relative_pt = self.get_field_param("relative_reference_point")
        dimensions = self.get_field_param("dimension_sizes")
        abs_posx = relative_pt[0] * dimensions[0]
        abs_posy = relative_pt[1] * dimensions[1]
        abs_posz = relative_pt[2] * dimensions[2]
        self._set_absolute_reference(np.array([abs_posx, abs_posy, abs_posz]))

    def get_field_param(self, parameter: str):
        """
        Get a field parameter by name.

        Parameters
        ----------
        parameter : str
            Name of the field parameter.

        Returns
        -------
        Any
            Value of the requested field parameter.
        """
        return self.params[parameter]

    def change_number_of_molecules(self, n: int):
        """
        Change the number of molecules in the field and generate new random positions.

        Parameters
        ----------
        n : int
            Number of molecules.
        """
        self.set_molecule_param("nMolecules", n)
        self.generate_random_positions()

    # molecule parameters
    def get_molecule_param(self, parameter: str):
        """
        Get a molecule parameter by name.

        Parameters
        ----------
        parameter : str
            Name of the molecule parameter.

        Returns
        -------
        Any
            Value of the requested molecule parameter.
        """
        return self.molecules_params[parameter]

    def set_molecule_param(self, parameter: str, value):
        """
        Set a molecule parameter.

        Parameters
        ----------
        parameter : str
            Name of the molecule parameter.
        value : Any
            Value to set.
        """
        if parameter == "orientation_per_particle":
            # check if the length of the list has the same size as number of particles
            if len(value) == self.molecules_params["nMolecules"]:
                self.molecules_params["orientations"] = value
            else:
                print("Number of values provided does not equal the number of existing particles")
        if parameter == "rotation_per_particle":
            # check if the length of the list has the same size as number of particles
            if len(value) == self.molecules_params["nMolecules"]:
                self.molecules_params["rotations"] = value
            else:
                print("Number of values provided does not equal the number of existing particles")
        else:
            self.molecules_params[parameter] = value

    def set_molecules_params(
        self,
        nMolecules: int,
        random_positions,
        random_orientations,
        random_rotations,
        **kwargs,
    ):
        """
        Set multiple molecule parameters at once.

        Parameters
        ----------
        nMolecules : int
            Number of molecules.
        random_positions : bool
            Whether to use random positions.
        random_orientations : bool
            Whether to use random orientations.
        random_rotations : bool
            Whether to use random rotations.
        **kwargs
            Additional molecule parameters.
        """
        # self.change_number_of_molecules(nMolecules)
        self.set_molecule_param("nMolecules", nMolecules)
        if random_positions:
            self.generate_random_positions()
        if random_orientations:
            self.generate_random_orientations()
        if random_rotations:
            self.initialise_random_rotations()
        for key, value in kwargs.items():
            self.molecules_params[key] = value

    def show_params(self):
        """
        Print the current field and molecule parameters.
        """
        print("Field parameters")
        for key, val in self.params.items():
            print(key, ": ", val)
        print("Molecule objects parameters")
        for key, val in self.molecules_params.items():
            print(key, ": ", val)

    # methods to prime molecule positions, orientations...
    def generate_random_positions(self):
        """
        Generate random positions for molecules within the field, optionally enforcing a minimal distance.
        """
        npositions = self.get_molecule_param("nMolecules")
        print(f"nmolecules: {npositions}")
        # print(f"Total positions: {npositions}")
        # this can only work after the size of field has been established
        if self.molecules_params["minimal_distance"] is not None:
            print(
                f"distributing with minimal distance: {self.molecules_params['minimal_distance']}"
            )
            self._random_pos_minimal_dist(npositions)
        else:
            print("Generating unconstrained random positions")
            xrel = np.random.uniform(size=npositions)
            yrel = np.random.uniform(size=npositions)
            zrel = np.random.uniform(size=npositions)
            rand = list(zip(xrel, yrel, zrel))
            self.molecules_params["relative_positions"] = rand
            self._gen_abs_from_rel_positions()

    def randomise_axial_position(self):
        rng = np.random.default_rng()
        nmolecules = self.get_molecule_param("nMolecules")
        axial_offsets = rng.choice(self.axial_offset, nmolecules, replace=True)
        for i, zpos in enumerate(axial_offsets):
            self.molecules_params["absolute_positions"][i][2] = zpos + self.z_offset


    def generate_random_orientations(self):
        """
        Generate random orientations for all molecules in the field.
        """
        # give new orientation
        norientations = self.get_molecule_param("nMolecules")
        orientations = []
        unconstrained = True
        rng = np.random.default_rng()
        xy_orientation_changes = np.zeros((norientations,))
        if self.xy_orientations is not None:
            xy_orientation_changes =  rng.choice(self.xy_orientations, norientations, replace=True)
            unconstrained = False
        if self.xz_orientations is not None:
            xz_orientation_changes =  rng.choice(self.xz_orientations, norientations, replace=True)
            unconstrained = False
        else:
            xz_orientation_changes = np.zeros((norientations,))
        if self.yz_orientations is not None:
            yz_orientation_changes =  rng.choice(self.yz_orientations, norientations, replace=True)
            unconstrained = False
        else:
            yz_orientation_changes = np.zeros((norientations,))
        if unconstrained:
            print("generating unconstrained randomised axis")
            # unconstrained randomisation
            for i in range(norientations):
                orientations.append(np.array(sampl.sample_spherical_normalised(1, ndim=3)))
            self.set_molecule_param("orientations", orientations)
        else:
            orientations_planewise = []
            for xy_xz_yz in zip(xy_orientation_changes, xz_orientation_changes, yz_orientation_changes):
                orientations_planewise.append(xy_xz_yz)
            self.set_molecule_param("orientations_planewise", orientations_planewise)
            self.set_molecule_param("orientations", None)

    def initialise_random_rotations(self, rotation_angles: list = None):
        """
        Generate random rotations around central axis for all molecules in the field.
        """
        nrotations = self.get_molecule_param("nMolecules")
        rng = np.random.default_rng()
        if self.rotation_angles is None:
            print("random unconstrained rotations")
            rotations = rng.integers(360, size=nrotations) 
        else:
            print("random rotations, from list")
            rotations = rng.choice(self.rotation_angles, nrotations, replace=True)
        self.set_molecule_param("rotations", rotations)

    def generate_global_orientation(self, global_orientation=None):
        """
        Set a global orientation for all molecules.

        Parameters
        ----------
        global_orientation : array-like, optional
            Orientation vector to assign to all molecules.
        """
        # give new orientation
        if global_orientation is not None:
            norientations = self.get_molecule_param("nMolecules")
            orientations = []
            for i in range(norientations):
                orientations.append(global_orientation)
            self.set_molecule_param("orientations", orientations)

    def _set_molecule_minimal_distance(self, dist):
        if dist > 0:
            print(
                f"minimal distance from particles set to {dist}. "
                f"FOV scale unit is {self.scale} meters"
            )
            self.molecules_params["minimal_distance"] = dist

    def _random_pos_minimal_dist(self, n):
        # convert minimal distance in relative units
        selected_positions = []
        selected_relative = []
        minimal_distance = self.get_molecule_param("minimal_distance")
        dimension_sizes = self.get_field_param("dimension_sizes")
        maxabs_posx = dimension_sizes[0]
        maxabs_posy = dimension_sizes[1]
        maxabs_posz = dimension_sizes[2]
        # print(f"Max values: {maxabs_posx},{maxabs_posy}, {maxabs_posz}")
        # sample the first point
        x = np.random.uniform(0, maxabs_posx, size=1)
        y = np.random.uniform(0, maxabs_posy, size=1)
        z = np.random.uniform(0, maxabs_posz, size=1)
        v0 = np.array([x, y, z]).reshape(-1)
        selected_positions.append(v0)
        flag = 0
        while len(selected_positions) < n and flag < 10000:
            flag = flag + 1
            x = np.random.uniform(0, maxabs_posx, size=1)
            y = np.random.uniform(0, maxabs_posy, size=1)
            z = np.random.uniform(0, maxabs_posz, size=1)
            new = np.array([x, y, z]).reshape(-1)
            is_available = 1
            for pos in range(len(selected_positions)):
                # verify is the next point is within the minimum distance
                # print(np.linalg.norm(new - selected_positions[pos]))
                if np.linalg.norm(new - selected_positions[pos]) < minimal_distance:
                    is_available = 0
                    # break
            if is_available:
                selected_positions.append(new)
        for abs_pos in selected_positions:
            rel_pos = abs_pos
            rel_pos[0] = abs_pos[0] / maxabs_posx
            rel_pos[1] = abs_pos[1] / maxabs_posy
            rel_pos[2] = abs_pos[2] / maxabs_posz
            selected_relative.append(rel_pos)
        # print(f"end flag : {flag}")
        self.molecules_params["relative_positions"] = selected_relative
        # self.set_molecule_param("absolute_positions", selected_positions)
        # self.absolute_pos = selected_positions

    def _calculate_absolute_position(self, relative_pos):
        fieldsizes = self.get_field_param("dimension_sizes")
        abs_posx = relative_pos[0] * fieldsizes[0]
        abs_posy = relative_pos[1] * fieldsizes[1]
        abs_posz = relative_pos[2] * fieldsizes[2]
        return np.array([abs_posx, abs_posy, abs_posz])

    def _gen_abs_from_rel_positions(self):
        abs_pos = []
        for pos in self.molecules_params["relative_positions"]:
            abs_pos.append(self._calculate_absolute_position(pos))
        self.set_molecule_param("absolute_positions", np.array(abs_pos))
        if self.axial_offset is not None:
            self.randomise_axial_position()

    # methods for fluorophores
    def _load_fluorophore_params(self, **fluodictionary):
        fname = fluodictionary["fluorophore_name"]
        fluo_plotting_params = dict(
            plotsize=20,
            plotalpha=1,
            plotmarker="o",
            plotcolour=fluodictionary["plotcolour"],
        )
        self.plotting_params[fname] = fluo_plotting_params
        self.fluoparams[fname] = dict(fluodictionary)
        # here add all other relevant parameters of a fluorophore in a separate atribute

    def _set_fluo_plotting_params(self, fluoname, plotcolour="m"):
        fluo_plotting_params = dict(
            plotsize=20, plotalpha=1, plotmarker="o", plotcolour=plotcolour
        )
        self.plotting_params[fluoname] = fluo_plotting_params

    def get_fluorophore_params(self):
        """
        Get the dictionary of fluorophore parameters.

        Returns
        -------
        dict
            Dictionary of fluorophore parameters.
        """
        return self.fluoparams

    # working with macromolecules
    def create_molecules_from_InstanceObject(self, InstancePrototype: LabeledInstance):
        """
        Create molecules in the field from a prototype instance.

        Parameters
        ----------
        InstancePrototype : LabeledInstance
            Prototype instance to copy for each molecule.
        """
        reps = self.get_molecule_param("nMolecules")
        particle_copy = copy.deepcopy(InstancePrototype)
        particle_copy.scale_coordinates_system(self.get_field_param("scale"))
        self.molecules_default_orientation = particle_copy.get_axis()
        if self.molecules_params["minimal_distance"] is None:
            self._set_molecule_minimal_distance(dist=particle_copy.radial_hindance)
        elif particle_copy.radial_hindance > self.molecules_params["minimal_distance"]:
            self._set_molecule_minimal_distance(dist=particle_copy.radial_hindance)
        #self._set_molecule_minimal_distance(dist=min_distance)
        if self.random_placing:
            self.generate_random_positions()
        self._gen_abs_from_rel_positions()
        # due to constraints, the actual number of particles might not be reps
        nmolecules = len(self.molecules_params["absolute_positions"])
        if reps > nmolecules:
            reps = nmolecules
            print(f"Updating Number of molecules: {reps}")
            self.set_molecule_param("nMolecules", reps)
        # prepare plotting params
        self.labels_plotting_params = dict(particle_copy.plotting_params)
        self.fluo2labels = dict(particle_copy.fluo2labels)
        self._set_plotting_params()
        molecules = []
        for r in range(reps):
            molecules.append(copy.deepcopy(particle_copy))
        self.molecules = molecules
        if self.sample_initial_orientation is not None:
            self.generate_global_orientation(self.sample_initial_orientation)
        if self.random_orientations:
            self.generate_random_orientations()
        if self.random_rotations:
            self.initialise_random_rotations()
            # self.relabel_molecules()
        self.relabel_molecules()

    def _create_instances_from_pdb(self, cif_f, structure_id, label_file):
        # this is exactly the same as the high level function
        structure_dictionary = {
            "file": cif_f,
            "identifier": structure_id,
            "format": "CIF",
        }
        with open(label_file, "r") as f:
            label_params = yaml.safe_load(f)
        # create the molecular structure
        Molecularstructure = MolecularReplicates(structure_dictionary)
        nlabels = len(label_params["labels"])
        # create labels
        label_handler = [Label() for i in range(nlabels)]
        for la in range(nlabels):
            label_handler[la].set_params(**label_params["labels"][la])
        # add labels to molecular structure
        for labelobject in label_handler:
            Molecularstructure.add_label(labelobject)
        inst_builder = Molecularstructure.create_instance_builder()
        particle = LabeledInstance()
        particle.load_source(**inst_builder)  # define the source
        particle.load_label_file(label_file)  # load the labelling entities
        particle.generate_instance()
        return particle

    def show_molecules(self):
        """
        Show all molecules in the field using their show_instance method.
        """
        if self.molecules is not None:
            for mol in self.molecules:
                mol.show_instance()
                mol.get_ref_point()

    def relabel_molecules(self):
        """
        Regenerate and rescale all molecules in the field.
        """
        if self.molecules:
            for mol in self.molecules:
                mol.generate_instance()
                # mol.show_instance()
                # mol.scale_coordinates_system(self.get_field_param("scale"))

    def relocate_molecules(self):
        """
        Move each molecule to its absolute position in the field.
        """
        if self.molecules:
            if self.get_molecule_param("absolute_positions") is not None:
                # move each molecule to their absolute position
                for mol, pos in zip(
                    self.molecules, self.get_molecule_param("absolute_positions")
                ):
                    mol.transform_translate(pos)

    def reorient_molecules(self):
        """
        Reorient each molecule in the field according to its assigned orientation.
        """
        if self.get_molecule_param("orientations") is not None:
            for mol, ori in zip(
                self.molecules, self.get_molecule_param("orientations")
            ):
                mol.transform_reorient_axis(ori)
        elif self.get_molecule_param("orientations_planewise") is not None:
            for mol, xy_xz_yz in zip(self.molecules, self.get_molecule_param("orientations_planewise")):
                mol.reorient_axis_by_plane(reset_orientation=True)
                mol.reorient_axis_by_plane(*xy_xz_yz, sequential=True)
        else:
            pass
            # print("molecule orientations has not been set. No reorientation done.")
    
    def rotate_molecules(self):
        """
        Rotate each molecule in the field according to its assigned angle of rotation.
        """
        if self.get_molecule_param("rotations") is not None:
            for mol, angle_degree in zip(
                self.molecules, self.get_molecule_param("rotations")
            ):
                mol.transform_rotate_around_axis(degree=angle_degree)

    def _set_plotting_params(self):
        for fluoname, labelname in self.fluo2labels.items():
            print(fluoname, labelname[0])
            fluo_colour = self.labels_plotting_params[labelname[0]]["plotcolour"]
            self._set_fluo_plotting_params(fluoname, fluo_colour)

    # CONSTRUCTION OF STATIC FIELD
    # methods for creating the coordinates emitters per fluorophore species
    # the rationale is, there will be emitters that correspond to one or other emitters
    # and all of them will be able to be visualised if the optics define it

    def construct_static_field(self, relocate=True, reorient=True, rotate=True):
        """
        Construct the static field by placing and orienting molecules and grouping emitters by fluorophore.

        Parameters
        ----------
        relocate : bool, optional
            If True, relocate molecules to their absolute positions. Default is True.
        reorient : bool, optional
            If True, reorient molecules. Default is True.
        """
        if self.molecules:
            # verify absolute positions exist
            if self.get_molecule_param("absolute_positions") is None:
                self._gen_abs_from_rel_positions()
            if self.params["absolute_reference_point"] is None:
                self.calculate_absolute_reference()
            # the only thing that will be needed every time is the placing of the particles
            # in the field
            if relocate:
                self.relocate_molecules()
            if reorient:
                self.reorient_molecules()
            if rotate:
                self.rotate_molecules()
            # pull emitters by fluorophores
            self._construct_channels_by_fluorophores()
        else:
            default_fluorophore = "AF647"
            # self.get_molecule_param("absolute_positions")
            self.fluorophre_emitters = {}
            self.fluorophre_emitters[default_fluorophore] = self.get_molecule_param(
                "absolute_positions"
            )

    def _construct_channels_by_fluorophores(self):
        """
        Group all emitter positions by fluorophore species.
        """
        nmolecules = self.get_molecule_param("nMolecules")
        if nmolecules == 0:
            print("Field contains no macromolecules")
        else:
            chs = {}
            if self.fluo2labels is not None:
                for fluoname in list(self.fluo2labels.keys()):
                    chs[fluoname] = self._pull_coordinates_per_fluorophore(fluoname)
                self.fluorophre_emitters = chs
            else:
                print("No relation from fluorophore to labels")

    def _pull_coordinates_per_fluorophore(self, fluophorename):
        """
        Get all emitter coordinates for a given fluorophore.

        Parameters
        ----------
        fluophorename : str
            Name of the fluorophore.

        Returns
        -------
        numpy.ndarray
            Array of emitter coordinates.
        """
        # first, verify that there are particles (Labelled instance)
        nmolecules = self.get_molecule_param("nMolecules")
        if nmolecules == 0:
            print("Field contains no particles")
        else:
            pulled = np.empty((0, 3))
            for mol in self.molecules:
                pulled = np.vstack(
                    (pulled, mol.get_emitters_by_fluorophore(fluophorename))
                )  # each particle is an object of LabeledInstance
            return pulled

    def get_emitters_by_fluorophores(self):
        """
        Get a dictionary of all emitters grouped by fluorophore.

        Returns
        -------
        dict
            Dictionary mapping fluorophore names to emitter coordinates.
        """
        if len(self.fluorophre_emitters) == 0:
            print("Constructing channels by fluorophore type")
            self._construct_channels_by_fluorophores()
        return dict(self.fluorophre_emitters)

    # methods for visualisation
    def show_field(
        self,
        fluo_type="all",
        view_init=[30, 0, 0],
        initial_pos=True,
        return_fig=False,
        axesoff=False,
        axis_object=None,
        zoom_in=None,
        emitters_plotsize=None,
        initial_pos_plotsize = 1,
        **kwargs,
    ):
        """
        Visualize the field and emitters in 3D.

        Parameters
        ----------
        fluo_type : str, optional
            Fluorophore type to display. Default is "all".
        view_init : list of int, optional
            Initial view angles [elev, azim, roll]. Default is [30, 0, 0].
        initial_pos : bool, optional
            If True, show initial positions. Default is True.
        return_fig : bool, optional
            If True, return the matplotlib figure. Default is False.

        Returns
        -------
        matplotlib.figure.Figure or None
            The figure if return_fig is True, otherwise None.
        """
        dimension_sizes = copy.copy(self.get_field_param("dimension_sizes"))
        if zoom_in and zoom_in > 0.2 and zoom_in < 1:
            x_to_remove = dimension_sizes[0] * zoom_in
            x_size = dimension_sizes[0] - x_to_remove
            y_to_remove = dimension_sizes[1] * zoom_in
            y_size = dimension_sizes[1] - y_to_remove
            xx, yy = np.meshgrid(range(int(x_size)), range(int(y_size)))
            xx = xx + (x_to_remove / 2)
            yy = yy + (y_to_remove / 2)
            zz = (yy) * 0
        else:
            xx, yy = np.meshgrid(
                range(int(dimension_sizes[0])), range(int(dimension_sizes[1]))
            )
            zz = (yy) * 0
        if axis_object is not None:
            ax = axis_object
            ax.plot_surface(xx, yy, zz, alpha=0.2, **kwargs)
        else:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
            ax.plot_surface(xx, yy, zz, alpha=0.2, **kwargs)

        if initial_pos:
            if self.molecules_params["absolute_positions"] is not None:
                add_ax_scatter(
                    ax,
                    format_coordinates(
                        np.array(self.molecules_params["absolute_positions"]),
                        plotsize=initial_pos_plotsize
                    ),
                )
        if len(self.fluorophre_emitters) == 0:
            print("Constructing static field")
            self.construct_static_field()
        if fluo_type == "all":
            print("Showing all fluorophores")
            for fname in self.fluorophre_emitters.keys():
                print(fname)
                if emitters_plotsize is not None:
                    self.plotting_params[fname]["plotsize"] = emitters_plotsize
                add_ax_scatter(
                    ax,
                    format_coordinates(
                        self.fluorophre_emitters[fname], **self.plotting_params[fname]
                    ),
                )
        else:
            if emitters_plotsize is not None:
                    self.plotting_params[fname]["plotsize"] = emitters_plotsize
            add_ax_scatter(
                ax,
                format_coordinates(
                    self.fluorophre_emitters[fluo_type],
                    **self.plotting_params[fluo_type],
                ),
            )
        ax.set_box_aspect(
            [ub - lb for lb, ub in (getattr(ax, f"get_{a}lim")() for a in "xyz")]
        )
        ax.view_init(elev=view_init[0], azim=view_init[1], roll=view_init[2])
        if axesoff:
            ax.set_axis_off()
        else:
            ax.set_xlabel("X (nm)")
            ax.set_ylabel("Y (nm)")
            ax.set_zlabel("Z (nm)")
        if axis_object is not None:
            return ax
        else:
            if return_fig:
                return fig
            else:
                fig.show()

    def expand_isotropically(self, factor):
        """
        Expand the field isotropically by a given factor.

        Parameters
        ----------
        factor : float
            Expansion factor (must be > 1).
        """
        if factor <= 1:
            print("Input factor has to be greater than 1")
        else:
            self.fov_absolute_ref_pt_xyself = self._calculate_absolute_position(
                self.fov_relative_ref_pt_xy
            )
            print(
                "Warining: this method changes the final emitter sizes"
                "without altering the scales. "
                "This is intended to be used as a last step"
            )
            for fname in self.fluorophre_emitters.keys():
                print(fname)
                print("expanding")
                # expanded[fname] = self.fluorophre_emitters[fname] * factor
                self.fluorophre_emitters[fname] = (
                    self.fluorophre_emitters[fname] * factor
                )
                print(self.fluorophre_emitters[fname])
            new_abs_ref = self.fov_absolute_ref_pt_xyself * factor
            print(f"newabs {new_abs_ref}")
            for fname in self.fluorophre_emitters.keys():
                print("adjusting")
                # expanded[fname] = self.fluorophre_emitters[fname] * factor
                self.fluorophre_emitters[fname], _ = transform_displace_set(
                    self.fluorophre_emitters[fname],
                    new_abs_ref,
                    self.fov_absolute_ref_pt_xyself,
                )
                print(self.fluorophre_emitters[fname])

    def export_field(self):
        """
        Export the field as a dictionary for use in imaging simulations.

        Returns
        -------
        dict
            Dictionary containing field emitters, scale, sizes, plotting parameters, and reference point.
        """
        if self.fluorophre_emitters is None:
            self.construct_static_field()
        field_emitters = self.fluorophre_emitters
        field_scale = self.get_field_param("scale")
        field_sizes = self.get_field_param("dimension_sizes")
        plotting_params = self.plotting_params
        reference_point = self.get_field_param("absolute_reference_point")
        export_field = dict(
            field_emitters=field_emitters,
            field_scale=field_scale,
            plotting_params=plotting_params,
            reference_point=reference_point,
            field_sizes=field_sizes,
            particle_positions=self.get_molecule_param("absolute_positions"),
            particle_size = self.molecules_params["minimal_distance"]
        )
        return export_field


def create_min_field(
    number_of_particles=1,
    random_placing=False,
    random_orientations=False,
    random_rotations=False,
    prints=False,
    **kwargs,
):
    """
    Create a minimal field with a specified number of particles.

    Parameters
    ----------
    number_of_particles : int, optional
        Number of particles to place in the field. Default is 1.
    random_placing : bool, optional
        If True, place particles randomly. Default is False.
    random_orientations : bool, optional
        If True, assign random orientations to particles. Default is False.
    prints : bool, optional
        If True, print debug information. Default is False.
    **kwargs
        Additional parameters for field creation.

    Returns
    -------
    Field
        The created Field object.
    """
    if prints:
        print("Initialising default field")
    coordinates_field = Field()
    # if molecule_pars:
    #    for key, value in molecule_pars.items():
    #        coordinates_field.molecules_params[key] = value
    coordinates_field.create_minimal_field(
        nmolecules=number_of_particles,
        random_placing=random_placing,
        random_orientations=random_orientations,
        random_rotations=random_rotations,
        **kwargs,
    )
    return coordinates_field


def gen_positions_from_image(
    img, mode="mask", pixelsize=None, min_distance=None, elevation_img=None, normalise_elevation=True, **kwargs
):
    """
    Generate relative positions from an image using either a mask or local maxima.

    Parameters
    ----------
    img : numpy.ndarray
        Input image.
    mode : str, optional
        Mode for position extraction ("mask" or "localmaxima"). Default is "mask".
    pixelsize : float, optional
        Size of a pixel.
    **kwargs
        Additional parameters for position extraction.

    Returns
    -------
    xyz_relative : list
        List of relative positions.
    image_physical_size : numpy.ndarray
        Physical size of the image.
    """
    if min_distance is None:
        min_distance = 1
    else:
        min_dist_pixels = min_distance / pixelsize
        min_distance = math.ceil(min_dist_pixels)
    npixels = list(img.shape)
    image_physical_size = np.zeros(shape=(2))
    image_physical_size[0] = pixelsize * npixels[0]
    image_physical_size[1] = pixelsize * npixels[1]
    if mode == "mask":
        if "npositions" not in kwargs.keys():
            npositions = 1
        else:
            npositions = kwargs["npositions"]
        pixel_positions = sampl.get_random_pixels(
            img, num_pixels=npositions, min_distance=min_distance
        )
    elif mode == "localmaxima":
        if "background" not in kwargs.keys():
            background = None
        else:
            background = kwargs.pop("background", None)
        if "sigma" not in kwargs.keys():
            sigma = None
        else:
            sigma = kwargs.pop("sigma", None)
        if "threshold" not in kwargs.keys():
            threshold = None
        else:
            threshold = kwargs.pop("threshold", None)
        pixel_positions, img_processed = metrics.local_maxima_positions(
            img,
            min_distance=min_distance,
            threshold=threshold,
            sigma=sigma,
            background=background,
            **kwargs
        )
    xyz_relative = metrics.pixel_positions_to_relative(
        pixel_positions, image_sizes=image_physical_size, pixelsize=pixelsize, elevation_img=elevation_img, normalise_elevation=normalise_elevation 
    )
    return xyz_relative, image_physical_size



import matplotlib.pyplot as plt
import seaborn as sns
from ..utils.transform.datatype import truncate
from matplotlib.ticker import FormatStrFormatter


def sns_heatmap_pivots(
    df_pivots,
    titles=None,
    conditions_cmaps=None,
    annotations=False,
    cmaps_range="same",
    figsize=[12, 10],
    return_figure=False,
    metric_name=None,
    decimals="%.4f",
    na_as_zero = True,
    **kwargs,
):
    conditions = list(df_pivots.keys())
    nconditions = len(conditions)
    if "annot_kws" not in kwargs.keys():
        annot_kws = {"size": 10, "rotation": 45}
    else:
        annot_kws = kwargs["annot_kws"]
    f, axes = plt.subplots(nconditions, 2, figsize=figsize, squeeze = False)
    plot_num = 0
    if cmaps_range == "same":
        hist_params = dict(vmin=0, vmax=1)
    elif cmaps_range == "each":
        hist_params = dict()
    if conditions_cmaps is None:
        conditions_cmaps = ["mako"] * nconditions
    if metric_name is not None:
        metric_n = metric_name
    else:
        metric_n = "Metric"
    for n, cond in enumerate(conditions):
        if na_as_zero:
            mask = df_pivots[cond][0].isna()
            df_pivots[cond][0][mask] = 0.0
            mask2 = df_pivots[cond][1].isna()
            df_pivots[cond][1][mask2] = 0.0
            #print("Na as 0")
            #print(df_pivots[cond][0])
        # mean
        sns.heatmap(
            df_pivots[cond][0],
            annot=annotations,
            annot_kws=annot_kws,
            ax=axes[n, 0],
            cmap=conditions_cmaps[n],
            # xticklabels=df_pivots[cond][0].columns.values.round(3),
            # yticklabels=df_pivots[cond][0].index.values.round(3),
            **hist_params,
            **kwargs
        )
        axes[n, 0].set_title(titles["category"] + ": " + cond + ". Mean " + metric_n)
        #axes[n, 0].yaxis.set_major_formatter(FormatStrFormatter(decimals))
        #axes[n, 0].xaxis.set_major_formatter(FormatStrFormatter(decimals))
        # std
        sns.heatmap(
            df_pivots[cond][1],
            annot=annotations,
            annot_kws=annot_kws,
            ax=axes[n, 1],
            cmap=conditions_cmaps[n],
            # xticklabels=df_pivots[cond][1].columns.values.round(3),
            # yticklabels=df_pivots[cond][1].index.values.round(3),
            **kwargs
        )
        axes[n, 1].set_title(
            titles["category"] + ": " + cond + ". Std Dev " + metric_n
        )
        #axes[n, 1].yaxis.set_major_formatter(FormatStrFormatter(decimals))
        #axes[n, 1].xaxis.set_major_formatter(FormatStrFormatter(decimals))

    f.tight_layout()

    if return_figure:
        plt.close()
        return f


def show_references(references):
    n_conditions = len(list(references.keys()))
    f, axes = plt.subplots(1, n_conditions, figsize=(12, 10))
    i = 0
    for cond, img in references.items():
        axes[i].imshow(img, cmap="grey")
        axes[i].set_title(f"Reference for: {cond}")
        i = i + 1


def show_example_test(
    queries, params, condition="STED_demo", replica_number=1, query_variant=0
):  #
    param_values = [truncate(p, 6) for p in params]
    print(param_values)
    combination_pars = [str(val) for val in param_values]
    print(combination_pars)
    combination_name = ",".join(combination_pars)
    print(combination_name)
    plt.imshow(
        queries[combination_name][replica_number][condition][query_variant], cmap="grey"
    )
    plt.title(combination_name)



import vlab4mic.generate.coordinates_field as field
from vlab4mic import workflows, experiments
from vlab4mic.utils import data_format
import numpy as np
import copy


def test_virtual_sample_params(experiment_7r5k_base):
    copy_exp = copy.deepcopy(experiment_7r5k_base)
    copy_exp.set_virtualsample_params(number_of_particles=2)
    copy_exp.build(modules=["virtualsample"])


def test_create_minimal_field():
    number_of_particles = 24
    test_field = field.create_min_field(
        number_of_particles=number_of_particles
    )
    test_field.molecules_params["nMolecules"] == number_of_particles
    test_field.change_number_of_molecules(25)
    #test_field.show_field()
    test_field.expand_isotropically(factor=2)


def test_gen_positions_from_image(experiment_7r5k_base):
    copy_exp = copy.deepcopy(experiment_7r5k_base)
    img_mask = np.random.rand(24, 24)
    p = 0.9
    img_mask[img_mask >= p] = 1
    img_mask[img_mask < p] = 0
    copy_exp.use_image_for_positioning(img_mask, mode="mask", pixelsize=100)


def test_epitopes_from_mask():
    probe_test = dict(
        probe_template = "Antibody",
        probe_name="custom",
        probe_target_type = "Sequence",
        probe_target_value = "EQATETQ",
        probe_distante_to_epitope = 0,
        probe_DoL=2,  
    )
    sample, experiment = experiments.generate_virtual_sample(
        structure = "1XI5",  ## Need to avoid
        probe_list=[probe_test,],
        clear_experiment=True,
        )
    img_mask = np.random.rand(24, 24)
    img_mask_elev = np.random.rand(24, 24)
    p = 0.9
    img_mask[img_mask >= p] = 1
    img_mask[img_mask < p] = 0
    experiment.use_image_for_positioning(
        img=img_mask,
        elevation_img=img_mask_elev,
        mode="mask",
        positions_are_epitopes=True,
        npositions=20,
        pixelsize=100,
        max_elevation_nm=100)
    assert len(list(experiment.probe_parameters.keys())) == 1
    assert len(list(experiment.particle.labels.keys())) == 1
    assert experiment.generators_status("structure") is False


def test_nparticles_constraints():
    number_of_particles = 200
    sample, expr = experiments.generate_virtual_sample(
        number_of_particles=number_of_particles,
        sample_dimensions=[200,200,100],
    )
    assert (
        expr.coordinate_field.molecules_params["minimal_distance"] is not None
    )
    assert len(expr.coordinate_field.molecules) < number_of_particles
    assert (
        expr.coordinate_field.molecules_params["nMolecules"]
        < number_of_particles
    )

def test_parameter_per_particle():
    relative_pos = [[0.24,0.24,0], [0.75,0.75,0]]
    rotation_per_particle = [0,30]
    orientation_per_particle = [
        [0,0,1],
        [1,0,0]
    ]
    vsample, myexperiment = experiments.generate_virtual_sample(
        particle_positions=relative_pos,
        rotation_per_particle=rotation_per_particle,
        orientation_per_particle=orientation_per_particle,
        clear_experiment=True,
    )
    assert myexperiment.coordinate_field.molecules[0].axis["direction"] == orientation_per_particle[0]
    assert myexperiment.coordinate_field.molecules[1].axis["direction"] == orientation_per_particle[1]


from skimage.metrics import structural_similarity as ssim
from scipy.ndimage import zoom
import numpy as np
from scipy.ndimage import gaussian_filter
from skimage.feature import peak_local_max
from scipy.stats import pearsonr
import cv2
import copy

def match_image_sizes(
        reference_image = None, 
        reference_image_pixelsize_nm = None,
        reference_image_mask = None,
        simulated_image = None,
        simulated_image_pixelsize_nm = None,
        simulated_image_mask = None
        ):
        union_mask = None
        if reference_image_pixelsize_nm and simulated_image_pixelsize_nm:
            reference_interpolated, simulated_image_interpolated = resize_images_interpolation(
                img1=reference_image,
                img2=simulated_image,
                px_size_im1=reference_image_pixelsize_nm,
                px_size_im2=simulated_image_pixelsize_nm,
            )
            if reference_image_mask is not None and simulated_image_mask is not None:
                reference_mask_interpolated, simulated_image_mask_interpolated = resize_images_interpolation(
                    img1=reference_image_mask,
                    img2=simulated_image_mask,
                    px_size_im1=reference_image_pixelsize_nm,
                    px_size_im2=simulated_image_pixelsize_nm,
                    interpolation_order=0 # becauese they are masks
                )
                union_mask = np.logical_or(
                    reference_mask_interpolated,
                    simulated_image_mask_interpolated)
                #masks_interpolated["reference_mask"] = reference_mask_interpolated
                #masks_interpolated["query_mask"] = simulated_image_mask_interpolated
                #masks_interpolated["union_mask"] = union_mask
        return reference_interpolated, simulated_image_interpolated, union_mask


def structural_similarity(
        reference_image = None, 
        reference_image_pixelsize_nm = None,
        reference_image_mask = None,
        simulated_image = None,
        simulated_image_pixelsize_nm = None,
        simulated_image_mask = None
):
    reference_interpolated, simulated_image_interpolated, union_mask = match_image_sizes(
        reference_image = reference_image, 
        reference_image_pixelsize_nm = reference_image_pixelsize_nm,
        reference_image_mask = reference_image_mask,
        simulated_image = simulated_image,
        simulated_image_pixelsize_nm = simulated_image_pixelsize_nm,
        simulated_image_mask = simulated_image_mask
    )
    similarity = ssim(
        reference_interpolated[union_mask],
        simulated_image_interpolated[union_mask],
        data_range=simulated_image_interpolated[union_mask].max() - simulated_image_interpolated[union_mask].min()
    )
    return similarity


def pearson_correlation(
        reference_image = None, 
        reference_image_pixelsize_nm = None,
        reference_image_mask = None,
        simulated_image = None,
        simulated_image_pixelsize_nm = None,
        simulated_image_mask = None
):
    reference_interpolated, simulated_image_interpolated, union_mask = match_image_sizes(
        reference_image = reference_image, 
        reference_image_pixelsize_nm = reference_image_pixelsize_nm,
        reference_image_mask = reference_image_mask,
        simulated_image = simulated_image,
        simulated_image_pixelsize_nm = simulated_image_pixelsize_nm,
        simulated_image_mask = simulated_image_mask
    )
    pearson_correlation, pval = pearsonr(
        reference_interpolated[union_mask].flatten(),
        simulated_image_interpolated[union_mask].flatten()
    )
    return pearson_correlation


def img_compare(ref, query, metric=["ssim",], force_match=False, zoom_in=0, ref_mask = None, query_mask = None, custom_metrics = None, **kwargs):
    """
    Compare two images using specified similarity metrics.

    Parameters
    ----------
    ref : numpy.ndarray
        Reference image.
    query : numpy.ndarray
        Image to compare against the reference.
    metric : list of str, optional
        List of metrics to use for comparison. Supported: "ssim", "pearson". Default is ["ssim"].
    force_match : bool, optional
        If True, resize images to match pixel sizes before comparison. Default is False.
    zoom_in : float, optional
        Zoom factor for cropping the images before comparison. Default is 0.
    **kwargs
        Additional keyword arguments, e.g., 'ref_pixelsize', 'modality_pixelsize'.

    Returns
    -------
    similarity_vector : list
        List of similarity values for each metric.
    ref : numpy.ndarray
        (Possibly resized) reference image.
    query : numpy.ndarray
        (Possibly resized) query image.
    """
    reference_image = ref.copy()
    query_image = query.copy()
    ref_pixelsize = None
    query_pixelsize = None
    if force_match:
        if 'ref_pixelsize' in kwargs and 'modality_pixelsize' in kwargs:
            ref_pixelsize = copy.copy(kwargs['ref_pixelsize'])
            query_pixelsize = copy.copy(kwargs['modality_pixelsize'])
            ref, query = resize_images_interpolation(
                img1=ref,
                img2=query,
                px_size_im1=kwargs['ref_pixelsize'],
                px_size_im2=kwargs['modality_pixelsize'],
                zoom_in=zoom_in
            )
            masks_used = dict()
            if ref_mask is not None and query_mask is not None:
                ref_mask_interpolated, query_mask_interpolated = resize_images_interpolation(
                    img1=ref_mask,
                    img2=query_mask,
                    px_size_im1=kwargs['ref_pixelsize'],
                    px_size_im2=kwargs['modality_pixelsize'],
                    zoom_in=zoom_in,
                    interpolation_order=0 # becauese they are masks
                )
                union_mask = np.logical_or(
                    ref_mask_interpolated,
                    query_mask_interpolated)
                masks_used["reference_mask"] = ref_mask_interpolated
                masks_used["query_mask"] = query_mask
                masks_used["union_mask"] = union_mask
        else:
            ref, query = resize_images_interpolation(
                img1=ref,
                img2=query,
                zoom_in=zoom_in
            )
    similarity_vector = []
    for method in metric:
        if method == "ssim":
            similarity = ssim(ref[union_mask], query[union_mask], data_range=query[union_mask].max() - query[union_mask].min())
            similarity_vector.append(similarity)
        elif method == "pearson":
            similarity, pval = pearsonr(ref[union_mask].flatten(), query[union_mask].flatten())
            similarity_vector.append(similarity)
        elif method in custom_metrics.keys():
            custom_measurement = custom_metrics[method](
                reference_image = reference_image,
                reference_image_pixelsize_nm = ref_pixelsize,
                simulated_image = query_image,
                simulated_image_pixelsize_nm = query_pixelsize,
                image_mask = union_mask,
                resized_reference_image = ref,
                resized_simulated_image = query,
                **kwargs
            )  
            #custom_measurement = metric_calculator.run_metric()
            if isinstance(custom_measurement, float):
                similarity_vector.append(custom_measurement)
            else:
                similarity_vector.append(None)
    return similarity_vector, ref, query, masks_used


def _padding(img1, img2, zoom_in=0, **kwargs):
    """
    Pad two images with zeros so they have the same shape, centering the originals.

    Parameters
    ----------
    img1 : numpy.ndarray
        First image.
    img2 : numpy.ndarray
        Second image.
    zoom_in : float, optional
        Zoom factor for cropping after padding. Default is 0.
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    tuple of numpy.ndarray
        The padded (and possibly zoomed) images.
    """
    if img1.shape == img2.shape:
        if zoom_in:
            img1 = zoom_img(img1, zoom_in=zoom_in)
            img2 = zoom_img(img2, zoom_in=zoom_in)
        return img1, img2
    else:
        height1, width1 = img1.shape
        height2, width2 = img2.shape
        max_width = max(width1, width2)
        max_height = max(height1, height2)

        # Calculate the padding needed for both images
        padding1_top = (max_height - height1) // 2
        padding1_left = (max_width - width1) // 2

        padding2_top = (max_height - height2) // 2
        padding2_left = (max_width - width2) // 2

        # Create padded arrays (with zero padding)
        img1_padded = np.zeros((max_height, max_width), dtype=np.float32)
        img2_padded = np.zeros((max_height, max_width), dtype=np.float32)

        # Place the original images into the center of the padded arrays
        img1_padded[
            padding1_top : padding1_top + height1,
            padding1_left : padding1_left + width1,
        ] = img1
        img2_padded[
            padding2_top : padding2_top + height2,
            padding2_left : padding2_left + width2,
        ] = img2
        if zoom_in:
            img1_padded = zoom_img(img1_padded, zoom_in=zoom_in)
            img2_padded = zoom_img(img2_padded, zoom_in=zoom_in)
        return img1_padded, img2_padded


def resize_images_interpolation(
    img1, img2, px_size_im1=1, px_size_im2=1, interpolation_order=3, zoom_in=0
):
    """
    Resize and interpolate images to match pixel sizes, then pad as needed.

    Parameters
    ----------
    img1 : numpy.ndarray
        Reference image.
    img2 : numpy.ndarray
        Image to compare against reference.
    px_size_im1 : float, optional
        Pixel size of image 1. Default is 1.
    px_size_im2 : float, optional
        Pixel size of image 2. Default is 1.
    interpolation_order : int, optional
        Order of interpolation (passed to scipy.ndimage.zoom). Default is 3.
    zoom_in : float, optional
        Zoom factor for cropping after resizing. Default is 0.

    Returns
    -------
    tuple of numpy.ndarray
        The resized and padded images.
    """
    pixel_size_ratio = px_size_im2 / px_size_im1

    resized_img2 = zoom(img2, pixel_size_ratio, order=interpolation_order)

    return _padding(img1, resized_img2, zoom_in)

def zoom_img(img, zoom_in=0, **kwargs):
    """
    Crop the center of the image by a zoom factor.

    Parameters
    ----------
    img : numpy.ndarray
        Image to crop.
    zoom_in : float, optional
        Fraction of the image to crop from each side. Default is 0.
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    numpy.ndarray
        Cropped image.
    """
    if zoom_in >= 0.9 or zoom_in < 0:
        zoom_in = 0
    if zoom_in:
        imshape = img.shape
        patchsize = np.ceil(imshape[0] - (imshape[0] * zoom_in))
        center_x = int(imshape[0] / 2)
        center_y = int(imshape[1] / 2)
        half_patch = int(patchsize / 2)
        img = img[center_y-half_patch:center_y+half_patch, center_x-half_patch:center_x+half_patch]
    return img


def image_preprocess(img, background=None, sigma=None, **kwargs):
    """
    Preprocess an image by subtracting background and applying Gaussian filter.

    Parameters
    ----------
    img : numpy.ndarray
        Image to preprocess.
    background : float or numpy.ndarray, optional
        Value or image to subtract as background.
    sigma : float, optional
        Standard deviation for Gaussian filter.
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    numpy.ndarray
        Preprocessed image.
    """
    if background:
        img = img - background
    if sigma:
        img = gaussian_filter(img, sigma=sigma)
    return img

def local_maxima_positions(img, min_distance=1, threshold=None, background=None, sigma=None, **kwargs):
    """
    Find local maxima positions in an image.

    Parameters
    ----------
    img : numpy.ndarray
        Input image.
    min_distance : int, optional
        Minimum number of pixels separating peaks. Default is 1.
    threshold : float, optional
        Minimum intensity of peaks. Default is None.
    **kwargs
        Additional keyword arguments for preprocessing.

    Returns
    -------
    xy : numpy.ndarray
        Array of (row, col) indices of local maxima.
    img_pre : numpy.ndarray
        Preprocessed image used for peak finding.
    """
    if len(img.shape) > 2:
        img = np.mean(img, axis=-1)
    # remove background as offset value
    img_pre = image_preprocess(img, background, sigma)
    xy = peak_local_max(img_pre, min_distance=min_distance, threshold_abs=threshold, **kwargs)
    return xy, img_pre

def pixel_positions_to_relative(indices, image_sizes, pixelsize, elevation_img, normalise_elevation=True):
    """
    Convert pixel indices to relative positions in the image.

    Parameters
    ----------
    indices : list or numpy.ndarray
        List of pixel indices.
    image_sizes : tuple or list
        Size of the image (height, width).
    pixelsize : float
        Size of a pixel.

    Returns
    -------
    xyz_relative : list of numpy.ndarray
        List of relative positions (x, y, z=0).
    """
    image_relative_positions=[(np.array(p)*pixelsize)/image_sizes for p in indices]
    if elevation_img is not None:
        xyz_relative = []
        if normalise_elevation:
            elevation_img_norm = elevation_img/elevation_img.max()
            for xypos, z_index in zip(image_relative_positions, indices):
                print(f"xy: {xypos},z_index: {z_index}, z: {elevation_img_norm[z_index[0], z_index[1]]}")
                xyz_relative.append(np.append(xypos, elevation_img_norm[z_index[0], z_index[1]]))
        else:
            for xypos, z_index in zip(image_relative_positions, indices):
                xyz_relative.append(np.append(xypos, elevation_img[z_index[0], z_index[1]]))
    else:
        xyz_relative = [ np.append(xypos, 0)  for xypos in image_relative_positions]
    return xyz_relative

def get_circles(img,
                blur_px=1,
                dp=0.1,
                minDist=3, 
                param1=1,
                param2=12,
                minRadius=3,
                maxRadius=7,
                **kwargs):
    """
    Detect circles in an image using the Hough Circle Transform.

    Parameters
    ----------
    img : numpy.ndarray
        Input image.
    blur_px : int, optional
        Size of the blur kernel. Default is 1.
    dp : float, optional
        Inverse ratio of the accumulator resolution to the image resolution. Default is 0.1.
    minDist : int, optional
        Minimum distance between detected centers. Default is 3.
    param1 : int, optional
        First method-specific parameter for HoughCircles. Default is 1.
    param2 : int, optional
        Second method-specific parameter for HoughCircles. Default is 12.
    minRadius : int, optional
        Minimum circle radius. Default is 3.
    maxRadius : int, optional
        Maximum circle radius. Default is 7.
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    circles : numpy.ndarray or None
        Detected circles.
    gray_blurred : numpy.ndarray
        Blurred grayscale image used for detection.
    cirlce_params : dict
        Parameters used for circle detection.
    """
    gray_blurred = img
    if gray_blurred.min() < 0:
        gray_blurred += -gray_blurred.min()
    if gray_blurred.dtype != np.uint8:
        if gray_blurred.dtype in [np.float32, np.float64]:
            #gray_blurred = np.uint8(gray_blurred * 255)
            gray_blurred = np.uint8(gray_blurred/gray_blurred.max() * 255)
        else:
            #gray_blurred = gray_blurred.astype(np.uint8)
            gray_blurred = (gray_blurred/256).astype('uint8')
    if blur_px:
        gray_blurred = cv2.blur(gray_blurred, (blur_px, blur_px), 0)
    else: 
        gray_blurred = gray_blurred
    circles = cv2.HoughCircles(
        gray_blurred,
        cv2.HOUGH_GRADIENT,
        dp=dp,
        minDist=minDist, 
        param1=param1,
        param2=param2,
        minRadius=minRadius,
        maxRadius=maxRadius,
    )
    cirlce_params=dict(
        dp=dp,
        minDist=minDist, 
        param1=param1,
        param2=param2,
        minRadius=minRadius,
        maxRadius=maxRadius,
    )
    return circles, gray_blurred, cirlce_params


import numpy as np


def builder_format(
    targets: dict,
    reference_point: np.ndarray,
    scale: float = 1e-10,
    axis: dict = dict(pivot=None, direction=None),
    info: str = "No info",
    **kwargs,
):
    """
    targets is a dictionary with at least two keys: "coordinates" and "normals"
    """
    instance_builder = dict(
        targets=targets,
        reference_point=reference_point,
        scale=scale,
        axis=axis,
        info=info,
    )
    for key, value in kwargs.items():
        instance_builder[key] = value
    return instance_builder


def label_builder_format(
    label_id,
    fluorophore_id,
    labelling_efficiency=1,
    target_info=None,
    as_linker=False,
    enable_wobble=False,
    wobble_theta = None,
    **kwargs,
):
    if target_info:
        label_info = dict(
            label_id=label_id,
            fluorophore_id=fluorophore_id,
            labelling_efficiency=labelling_efficiency,
            target_info=target_info,
            as_linker=as_linker,
            enable_wobble=enable_wobble,
            wobble_theta=wobble_theta
        )
        for key, val in kwargs.items():
            label_info[key] = val
    else:
        label_info = dict(
            label_id=label_id,
            fluorophore_id=fluorophore_id,
            labelling_efficiency=labelling_efficiency,
        )
    return label_info


def struct_params_format(struct_ID=None, title=None, labels=list(["<None>"])):
    structure_params = dict(id=struct_ID, title=title, labels=labels)
    return structure_params



from dataclasses import dataclass, field, fields
from typing import List, Dict
import numpy as np
import vlab4mic
from .workflows import (
    load_structure,
    create_imaging_system,
    particle_from_structure,
    field_from_particle,
    generate_multi_imaging_modalities,
)
from .utils.transform.points_transforms import apply_euler_rotation
from .generate import coordinates_field
from .utils.data_format.structural_format import label_builder_format
from .utils.data_format import configuration_format
import os
from vlab4mic.utils.io.yaml_functions import load_yaml
import numpy as np
import os
import copy
from vlab4mic.utils.io import yaml_functions
from IPython.utils import io
from pathlib import Path
import sys

IN_COLAB = "google.colab" in sys.modules
if IN_COLAB:
    output_path = "/content/vlab4mic_outputs"
else:
    output_path = Path.home() / "vlab4mic_outputs"


if not os.path.exists(output_path):
    os.makedirs(output_path)


@dataclass
class ExperimentParametrisation:
    experiment_id: str = "vLab4mic_experiment"
    structure_id: str = "1XI5"
    structure_format: str = "CIF"
    configuration_path: str = ""
    structure_label: str = "NHS_ester"
    fluorophore_id: str = "AF647"
    coordinate_field_id: str = "square1x1um_randomised"
    selected_mods: Dict[str, int] = field(default_factory=dict)
    imaging_modalities: Dict[str, int] = field(default_factory=dict)
    probe_parameters: Dict[str, int] = field(default_factory=dict)
    structural_integrity_eps: Dict[str, int] = field(default_factory=dict)
    sweep_pars: Dict[str, int] = field(default_factory=dict)
    objects_created: Dict[str, int] = field(default_factory=dict)
    output_directory: str = None
    example_structures = ["3J3Y", "7R5K", "1XI5", "8GMO"]
    example_modalities = ["Widefield", "Confocal", "AiryScan", "STED", "SMLM"]
    random_seed: int = None

    def __post_init__(self):
        self.output_directory = str(output_path)
        pck_dir = os.path.dirname(os.path.abspath(vlab4mic.__file__))
        local_dir = os.path.join(pck_dir, "configs")
        self.configuration_path = local_dir
        self.structure_path = None
        # keep track of objects created
        self.objects_created = dict(
            structure=False,
            particle=False,
            exported_coordinate_field=False,
            coordinate_field=False,
            imager=False,
            output_reference=False,
        )
        self.virtualsample_params = dict()
        self.structural_integrity_eps["structural_integrity"] = None
        self.structural_integrity_eps["eps1"] = None
        self.structural_integrity_eps["eps2"] = None
        self.structural_integrity_eps["use_structural_integrity"] = False
        # read information of local modalities configuration
        modalities_dir = os.path.join(local_dir, "modalities")
        modalities_names_list = []
        modality_parameters = {}
        self.config_modalities = dict()
        for mods in os.listdir(modalities_dir):
            if (
                os.path.splitext(mods)[-1] == ".yaml"
                and "_template" not in mods
            ):
                modalities_names_list.append(os.path.splitext(mods)[0])
        for mod in modalities_names_list:
            mod_info, mod_configuration = (
                configuration_format.compile_modality_parameters(
                    mod, local_dir
                )
            )
            modality_parameters[mod] = mod_info
            self.config_modalities[mod] = mod_configuration
        self.local_modalities_names = modalities_names_list
        self.local_modalities_parameters = modality_parameters
        probes_dir = os.path.join(local_dir, "probes")
        structure_dir = os.path.join(local_dir, "structures")
        self.config_probe_params = {}
        self.config_probe_models_names = []
        self.config_global_probes_names = []
        self.config_probe_per_structure_names = {}
        for p_file in os.listdir(probes_dir):
            if (
                os.path.splitext(p_file)[-1] == ".yaml"
                and "_template" not in p_file
            ):
                label_config_path = os.path.join(probes_dir, p_file)
                label_parmeters = vlab4mic.load_yaml(label_config_path)
                # print(label_parmeters)
                lablname = os.path.splitext(p_file)[0]
                if "Mock" in label_parmeters["known_targets"]:
                    self.config_probe_models_names.append(lablname)
                    self.config_probe_params[lablname] = label_parmeters
                elif "Generic" in label_parmeters["known_targets"]:
                    self.config_global_probes_names.append(lablname)
                    self.config_probe_params[lablname] = label_parmeters
                else:
                    self.config_probe_params[lablname] = label_parmeters
                    for struct in label_parmeters["known_targets"]:
                        if (
                            struct
                            in self.config_probe_per_structure_names.keys()
                        ):
                            self.config_probe_per_structure_names[
                                struct
                            ].append(lablname)
                        else:
                            self.config_probe_per_structure_names[struct] = [
                                lablname,
                            ]
        self.demo_structures = []
        # get available structure IDs
        self.structures_info_list = dict()
        structure_dir = os.path.join(self.configuration_path, "structures")
        fluorophores_dir = os.path.join(
            self.configuration_path, "fluorophores"
        )
        self.fluorophore_parameters = dict()
        for file in os.listdir(fluorophores_dir):
            if (
                os.path.splitext(file)[-1] == ".yaml"
                and "_template" not in file
            ):
                fluorophore_id = os.path.splitext(file)[0]
                self.fluorophore_parameters[fluorophore_id] = load_yaml(
                    os.path.join(fluorophores_dir, file)
                )
        self.fluorophore_parameters_template = load_yaml(
            os.path.join(fluorophores_dir, "_template_.yaml")
        )
        probes_dir = os.path.join(self.configuration_path, "probes")
        modalities_dir = os.path.join(self.configuration_path, "modalities")
        for file in os.listdir(structure_dir):
            if (
                os.path.splitext(file)[-1] == ".yaml"
                and "_template" not in file
            ):
                structure_params = load_yaml(os.path.join(structure_dir, file))
                struct_id = structure_params["model"]["ID"]
                if struct_id in self.example_structures:
                    strict_title = structure_params["model"]["title"]
                    id_title = struct_id + ": " + strict_title
                    self.structures_info_list[id_title] = struct_id
                    self.demo_structures.append(id_title)
        self.config_directories = dict(
            structure=structure_dir,
            fluorophores=fluorophores_dir,
            probes=probes_dir,
            modalities=modalities_dir,
            base=self.configuration_path,
        )
        param_settings_file = os.path.join(
            self.config_directories["base"], "parameter_settings.yaml"
        )
        self.param_settings = yaml_functions.load_yaml(param_settings_file)
        self.results = dict()
        self.create_example_experiment()
        self.modality_noise_images = dict()
        if self.random_seed is not None:
            np.random.seed(self.random_seed)

    def select_structure(self, structure_id="1XI5", build=True, structure_path:str = None):
        """
        Select a molecular structure by its identifier and optionally build the structure module.

        Parameters
        ----------
        :param structure_id : str, optional
            The identifier of the structure to select. Default is "1XI5".
        :param build : bool, optional
            If True, triggers the build process for the structure module. Default is True.

        Returns
        -------
        None
        """
        if structure_path is not None:
            self.structure_path = structure_path
            self.structure_id = structure_id
            self.structure_format = structure_path.split(".")[-1].upper()
        else:
            self.structure_path = None
            self.structure_id = structure_id
        if build:
            self.build(modules=["structure"])

    def clear_structure(self):
        """
        Clear the current structure by resetting the structure ID and related parameters.
        This method sets the structure ID to None and resets the structure object if it exists.
        Returns
        -------
        None
        """
        self.structure_id = None
        self.structure_path = None
        if self.generators_status("structure"):
            self.structure = None
        self.objects_created["structure"] = False
        print("Structure cleared")

    def add_modality(self, modality_name, save=False, modality_template=None, **kwargs):
        """
        Add a new imaging modality to the experiment.

        If the specified modality name exists in the local modalities, it copies the parameters from the local template.
        Otherwise, it uses the 'Widefield' modality parameters as a template to create a new modality.
        Additional parameters for the modality can be specified via keyword arguments and will override the defaults.
        Optionally, the modality output can be saved by setting `save` to True.

        Parameters
        ----------
        :param modality_name : str
            The name of the modality to add.
        :param save : bool, optional
            If True, saves the modality output. Default is False.
        :param **kwargs
            Arbitrary keyword arguments to override or set specific modality parameters.

        Returns
        -------
        None
        """
        if modality_name in self.local_modalities_names:
            self.imaging_modalities[modality_name] = copy.deepcopy(
                self.local_modalities_parameters[modality_name]
            )
            for param, value in kwargs.items():
                self.imaging_modalities[modality_name][param] = value
            self.set_modality_acq(modality_name, save=save)
        else:
            print(
                f"Modality {modality_name} not found in demo modalities. Using template to create new."
            )
            if modality_template is not None:
                template_modality = modality_template
            else:
                template_modality = "Widefield"
            print(f"Using {template_modality} as template.")
            self.imaging_modalities[modality_name] = copy.deepcopy(
                self.local_modalities_parameters[template_modality]
            )
            self.imaging_modalities[modality_name]["modality"] = modality_name
            for param, value in kwargs.items():
                self.imaging_modalities[modality_name][param] = value
            self.set_modality_acq(modality_name, save=save)
            self.local_modalities_names.append(modality_name)
            self.local_modalities_parameters[modality_name] = copy.deepcopy(
                self.imaging_modalities[modality_name]
            )

    def update_modality(
        self,
        modality_name,
        pixelsize_nm: int = None,
        lateral_resolution_nm: int = None,
        axial_resolution_nm: int = None,
        psf_voxel_nm: int = None,
        depth_of_field_nm: int = None,
        remove=False,
        **kwargs,
    ):
        """
        Update or remove an imaging modality's parameters.

        This method allows updating specific parameters of an imaging modality, such as pixel size, lateral and axial resolution, and PSF voxel size.
        If `remove` is True, the modality is removed from the internal dictionaries.

        Parameters
        ----------
        :param modality_name : str
            The name of the imaging modality to update or remove.
        :param pixelsize_nm : int, optional
            The new pixel size in nanometers. If provided, updates the detector pixel size.
        :param lateral_resolution_nm : int, optional
            The new lateral resolution in nanometers. If provided, updates the lateral standard deviations of the PSF.
        :param axial_resolution_nm : int, optional
            The new axial resolution in nanometers. If provided, updates the axial standard deviation of the PSF.
        :param psf_voxel_nm : int, optional
            The new PSF voxel size in nanometers. If provided, updates the PSF voxel size for all axes.
        :param remove : bool, optional
            If True, removes the specified modality from the internal dictionaries. Default is False.

        Notes
        -----
        After updating any parameters, the imaging modality is reconfigured in the imager.
        """
        if remove:
            self.imaging_modalities.pop(modality_name, None)
            self.selected_mods.pop(modality_name, None)
        else:
            changes = False
            if pixelsize_nm is not None:
                self.imaging_modalities[modality_name]["detector"][
                    "pixelsize"
                ] = (pixelsize_nm / 1000)
                changes = True
            if lateral_resolution_nm is not None:
                voxel_size = self.imaging_modalities[modality_name][
                    "psf_params"
                ]["voxelsize"][0]
                self.imaging_modalities[modality_name]["psf_params"][
                    "std_devs"
                ][0] = (lateral_resolution_nm / voxel_size)
                self.imaging_modalities[modality_name]["psf_params"][
                    "std_devs"
                ][1] = (lateral_resolution_nm / voxel_size)
                changes = True
            if axial_resolution_nm is not None:
                voxel_size = self.imaging_modalities[modality_name][
                    "psf_params"
                ]["voxelsize"][0]
                self.imaging_modalities[modality_name]["psf_params"][
                    "std_devs"
                ][2] = (axial_resolution_nm / voxel_size)
                changes = True
            if psf_voxel_nm is not None:
                self.imaging_modalities[modality_name]["psf_params"][
                    "voxelsize"
                ] = [
                    psf_voxel_nm,
                    psf_voxel_nm,
                    psf_voxel_nm,
                ]
                changes = True
            if depth_of_field_nm is not None:
                depth_in_slices = None
                voxel_size = self.imaging_modalities[modality_name][
                    "psf_params"
                ]["voxelsize"][2]
                depth = int(depth_of_field_nm / voxel_size)
                self.imaging_modalities[modality_name]["psf_params"][
                    "depth"
                ] = depth
                changes = True
            if changes:
                self.imager.set_imaging_modality(
                    **self.imaging_modalities[modality_name]
                )

    def set_modality_acq(
        self,
        modality_name,
        exp_time=0.001,
        noise=True,
        save=False,
        nframes=1,
        channels=[
            "ch0",
        ],
        **kwargs,
    ):
        """
        Configure and set acquisition parameters for a specified imaging modality.

        Parameters
        ----------
        :param modality_name : str
            The name of the imaging modality to configure.
        :param exp_time : float, optional
            Exposure time for the acquisition in seconds. Default is 0.001.
        :param noise : bool, optional
            Whether to include noise in the acquisition. Default is True.
        :param save : bool, optional
            Whether to save the acquired data. Default is False.
        :param nframes : int, optional
            Number of frames to acquire. Default is 1.
        :param channels : list of str, optional
            List of channel names to use. Default is ["ch0"].
        :param **kwargs
            Additional keyword arguments for modality-specific parameters.

        Notes
        -----
        If the specified modality is not available in `self.imaging_modalities`, a message is printed and no action is taken.
        Acquisition parameters are formatted using `configuration_format.format_modality_acquisition_params` and stored in `self.selected_mods`.
        """
        if modality_name in self.imaging_modalities.keys():
            self.selected_mods[modality_name] = (
                configuration_format.format_modality_acquisition_params(
                    exp_time, noise, save, nframes, channels
                )
            )
        else:
            print("Modality not selected")

    def clear_modalities(self):
        """
        Clear all selected imaging modalities and their acquisition parameters.
        This method resets the `imaging_modalities` and `selected_mods` dictionaries to empty states,
        effectively removing all configured modalities from the experiment.
        Returns
        -------
        None
        """
        modality_names = list(self.imaging_modalities.keys())
        for modality_name in modality_names:
            self.update_modality(modality_name, remove=True)
        if self.generators_status("imager"):
            self.imager = None
            self.objects_created["imager"] = False
        print("All modalities cleared")

    def reset_to_defaults(self, module="acquisitions", **kwargs):
        """
        Reset acquisition parameters or other module settings to their default values.

        Parameters
        ----------
        :param module : str, optional
            The module to reset. Default is "acquisitions".
        :param **kwargs
            Additional keyword arguments passed to the reset function.

        Returns
        -------
        None
        """
        if module == "acquisitions":
            for mod_name in self.selected_mods.keys():
                self.set_modality_acq(mod_name, **kwargs)

    def _build_structure(self, keep=True):
        """
        Build the structure object for the experiment.

        Parameters
        ----------
        :param keep : bool, optional
            If True, store the structure in the experiment. Default is True.

        Returns
        -------
        None
        """
        if self.structure_id and self.structure_path:
            struct, struct_param = load_structure(
                self.structure_id, self.configuration_path, self.structure_path, self.structure_format
            )
            self.structure = struct
            self.objects_created["structure"] = True
        elif self.structure_id:
            struct, struct_param = load_structure(
                self.structure_id, self.configuration_path
            )
            self.structure = struct
            self.objects_created["structure"] = True
        else:
            print("No structure ID")

    def _build_label(self, lab_eff=1, keep=False, **kwargs):
        """
        Create a list with one dictionary with the setup info for creating a label.

        Assumes there is only one label stored in the attribute "structure_label" and one in "fluorophore_id".

        Parameters
        ----------
        :param lab_eff : float, optional
            Labelling efficiency. Default is 1.
        :param keep : bool, optional
            If True, keep the label. Default is False.
        :param **kwargs
            Additional keyword arguments for label creation.

        Returns
        -------
        list
            List of label dictionaries.
        """
        labels_list = []
        for probe_name, probe_params in self.probe_parameters.items():
            labels_list.append(
                copy.deepcopy(self.probe_parameters[probe_name])
                #
            )
        return labels_list

    def _check_if_structural_integrity(self):
        """
        Check if structural_integrity parameters are set and update the use_structural_integrity flag.

        Returns
        -------
        None
        """
        if (
            self.structural_integrity_eps["structural_integrity"] is not None
            and self.structural_integrity_eps["eps1"]
            and self.structural_integrity_eps["eps2"]
        ):
            self.structural_integrity_eps["use_structural_integrity"] = True
        else:
            self.structural_integrity_eps["use_structural_integrity"] = False

    def _build_particle(self, lab_eff=1.0, structural_integrity_build=None, keep=False):
        """
        Build the particle object for the experiment, optionally adding structural_integritys.

        Parameters
        ----------
        :param lab_eff : float, optional
            Labelling efficiency. Default is 1.0.
        :param structural_integrity_build : float or None, optional
            Defect parameter to use. Default is None.
        :param keep : bool, optional
            If True, store the particle in the experiment. Default is False.

        Returns
        -------
        particle or None
            The particle object if created, else None.
        """
        if self.generators_status("structure"):
            self._check_if_structural_integrity()
            labels_list = self._build_label(lab_eff=lab_eff)
            if len(labels_list) > 0:
                particle, label_params_list = particle_from_structure(
                    self.structure, labels_list, self.configuration_path
                )
                if self.structural_integrity_eps["use_structural_integrity"]:
                    if structural_integrity_build is not None:
                        structural_integrity = structural_integrity_build
                    else:
                        structural_integrity = self.structural_integrity_eps["structural_integrity"]
                    particle.add_structural_integrity(
                        eps1=self.structural_integrity_eps["eps1"],
                        xmer_neigh_distance=self.structural_integrity_eps["eps2"],
                        integrity=structural_integrity,
                    )
                else:
                    particle.add_structural_integrity(
                        integrity=1,
                    )
                if keep:
                    self.particle = particle
                    self.objects_created["particle"] = True
                return particle

    def _build_coordinate_field(
        self,
        use_self_particle=True,
        keep=False,
        coordinate_field_path=None,
        **kwargs,
    ):
        """
        Build the coordinate field for the experiment, either from the current particle or as a minimal field.

        Parameters
        ----------
        :param use_self_particle : bool, optional
            If True, use the current particle to build the field. Default is True.
        :param keep : bool, optional
            If True, store the field in the experiment. Default is False.
        :param coordinate_field_path : str or None, optional
            Path to a coordinate field file. Default is None.
        :param **kwargs
            Additional keyword arguments for field creation.

        Returns
        -------
        exported_field or None
            The exported field if created, else None.
        """
        self.exported_coordinate_field = None
        if use_self_particle and self.generators_status("particle"):
            print("creating field from existing particle")
            exported_field, fieldobject = field_from_particle(
                self.particle, **self.virtualsample_params, **kwargs
            )
            self.virtualsample_params["minimal_distance"] = (
                fieldobject.molecules_params["minimal_distance"]
            )
            if keep:
                self.exported_coordinate_field = exported_field
                self.objects_created["exported_coordinate_field"] = True
                self.coordinate_field = fieldobject
                self.objects_created["coordinate_field"] = True
            return exported_field
        else:
            # create minimal field
            fieldobject = coordinates_field.create_min_field(
                **self.virtualsample_params, **kwargs
            )
            exported_field = fieldobject.export_field()
            if keep:
                self.exported_coordinate_field = exported_field
                self.objects_created["exported_coordinate_field"] = True
                self.coordinate_field = fieldobject
                self.objects_created["coordinate_field"] = True

    def _build_imager(self, use_local_field=False, prints=True):
        """
        Build the imager object for the experiment using the selected imaging modalities.

        Parameters
        ----------
        :param use_local_field : bool, optional
            If True, use the local exported coordinate field. Default is False.
        :param prints : bool, optional
            If True, print status messages. Default is True.

        Returns
        -------
        None
        """
        if self.imaging_modalities:
            # print(f"Using selected mods: {self.imaging_modalities.keys()}")
            mods_list = list(self.imaging_modalities.keys())
            if use_local_field and self.generators_status(
                "exported_coordinate_field"
            ):
                self.imager, modality_parameters = create_imaging_system(
                    exported_field=self.exported_coordinate_field,
                    modalities_id_list=mods_list,
                    mod_params=self.imaging_modalities,
                    config_dir=self.configuration_path,
                    fluorophore_parameters=self.fluorophore_parameters,
                )
            else:
                if prints:
                    print(
                        "Local field missing or unused. Creating imager without particles"
                    )
                self.imager, modality_parameters = create_imaging_system(
                    modalities_id_list=mods_list,
                    mod_params=self.imaging_modalities,
                    config_dir=self.configuration_path,
                    fluorophore_parameters=self.fluorophore_parameters
                )
            # update the base acquisition parameters to consider all channels
            imager_channels = []
            anymod = list(self.imager.modalities.keys())[0]
            for chann in self.imager.modalities[anymod]["filters"].keys():
                imager_channels.append(chann)
            nchannels = len(imager_channels)
            self.reset_to_defaults(module="acquisitions", channels=imager_channels)
            self.objects_created["imager"] = True
        else:
            self.imager = None
            print("No modalities")

    def _gen_modality_noise_images(self):
        for modality_name in self.imager.modalities.keys():
            self.modality_noise_images[modality_name], _beads, _im_noiseless, _b_noiseless = self.imager.generate_imaging(
                modality=modality_name,
                noise=True, exp_time=0)

    def generators_status(self, generator_name):
        """
        Return the status of the specified generator.

        Parameters
        ----------
        :param generator_name : str
            The name of the generator whose status is to be retrieved.

        Returns
        -------
        object
            The status or object associated with the given generator name from the objects_created dictionary.

        Raises
        ------
        KeyError
            If the specified generator_name does not exist in objects_created.
        """
        return self.objects_created[generator_name]

    def build(
        self,
        use_locals=True,
        modules: list = [
            "all",
        ],
        **kwargs,
    ):
        """
        Build various simulation objects based on the specified modules.

        Parameters
        ----------
        :param use_locals : bool, optional
            Determines whether to use local variables or attributes when building objects. Default is True.
        :param modules : list, optional
            List of module names to build. If "all" is included, builds all available modules: "structure", "particle", "coordinate_field", and "imager". Default is ["all"].
        :param **kwargs
            Additional keyword arguments for building modules.

        Notes
        -----
        The order of building is: structure, particle, coordinate_field, imager.
        """
        print("Building objects")
        if "all" in modules:
            build_list = [
                "structure",
                "particle",
                "coordinate_field",
                "imager",
            ]
        else:
            build_list = modules
        if "structure" in build_list:
            self._build_structure()
        if "particle" in build_list:
            self._build_particle(keep=use_locals)
        if "coordinate_field" in build_list:
            self._build_coordinate_field(
                use_self_particle=use_locals, keep=use_locals
            )
        if "imager" in build_list:
            self._build_imager(use_local_field=use_locals)

    def _gen_reference(
        self, write=False, keep=False, ref_acq_pars=None, modality_wise=False
    ):
        """
        Calculate a reference image of the virtual sample by using the ideal parameters for each of the parameters to sweep.

        Requires the dictionary of the params2sweep. If modality_wise is True (default), a reference is calculated for each modality with that modality's parameters. Otherwise, it will use the Reference modality configuration to generate an idealised image as reference.

        Parameters
        ----------
        :param write : bool, optional
            Whether to write the reference image to disk. Default is False.
        :param keep : bool, optional
            If True, store the reference in the experiment. Default is False.
        :param ref_acq_pars : dict or None, optional
            Acquisition parameters for the reference. Default is None.
        :param modality_wise : bool, optional
            If True, calculate reference for each modality. Default is False.

        Returns
        -------
        tuple
            (_reference, _reference_parameters)
        """
        reference_pars = dict()
        output_name = "REFERENCE_"
        # get ideal parameters for virtual sample
        for param_name, param_settings in self.sweep_pars.items():
            reference_pars[param_name] = param_settings["ideal"]
        print(reference_pars)
        # generate particle
        tmp_particle = self._build_particle(
            lab_eff=reference_pars["labelling_efficiency"], keep=True
        )
        # use particle to create new field
        tmp_exported_field = self._build_coordinate_field()
        # create ideal image modality
        if modality_wise:
            self.imager.import_field(**tmp_exported_field)
            _reference = generate_multi_imaging_modalities(
                image_generator=self.imager,
                experiment_name=output_name,
                savingdir=self.output_directory,
                write=write,
            )
        else:
            reference_imager, ref_modality_parameters = create_imaging_system(
                modalities_id_list=["Reference"],
                config_dir=self.configuration_path,
                fluorophore_parameters=self.fluorophore_parameters
            )
            reference_imager.import_field(**tmp_exported_field)
            # make a copy
            _reference = dict()
            _reference_parameters = dict()
            for mod_name in list(self.imager.modalities.keys()):
                _reference_iteration = generate_multi_imaging_modalities(
                    image_generator=reference_imager,
                    experiment_name=output_name,
                    acquisition_param=ref_acq_pars,
                    savingdir=self.output_directory,
                    write=write,
                )
                _reference[mod_name] = _reference_iteration["Reference"]
                imager_scale = reference_imager.roi_params["scale"]
                scalefactor = np.ceil(
                    imager_scale / 1e-9
                )  # resulting pixel size in nanometers
                _reference_parameters[mod_name] = dict(
                    ref_pixelsize=reference_imager.modalities["Reference"][
                        "detector"
                    ]["pixelsize"]
                    * scalefactor
                )
        if keep:
            self.experiment_reference = _reference
            self.objects_created["output_reference"] = True
        return _reference, _reference_parameters

    def clear_labelled_structure(self):
        self.remove_probes()

    def clear_virtual_sample(self):
        if self.generators_status("coordinate_field"):
            self.exported_coordinate_field = None
            self.objects_created["exported_coordinate_field"] = False
        print("Virtual sample cleared")

    def create_example_experiment(self):
        with io.capture_output() as captured:
            self.select_structure("1XI5", build=False)
            self.add_probe()
            self.set_virtualsample_params()
            for modality_name in self.example_modalities:
                self.add_modality(modality_name)
            self.build()
        print("Experiment created with default parameters")

    def clear_experiment(self):
        """
        Clear the current experiment by resetting all parameters and objects.
        """
        self.clear_structure()
        self.clear_labelled_structure()
        self.clear_virtual_sample()
        self.clear_modalities()
        self.results = dict()

    def run_simulation(
        self,
        name="vlab4mic_experiment",
        acq_params=None,
        save=False,
        modality="All",
        **kwargs,
    ):
        """
        Run a simulation for the specified imaging modality or all modalities.

        Parameters
        ----------
        :param name : str, optional
            Name of the experiment or simulation. Default is "vlab4mic_experiment".
        :param acq_params : dict or None, optional
            Acquisition parameters for the simulation. If None and modality is "All", uses self.selected_mods.
        :param save : bool, optional
            Whether to save the simulation output to disk. Default is False.
        :param modality : str, optional
            The imaging modality to simulate. If "All", simulates all available modalities. Default is "All".
        :param **kwargs
            Additional keyword arguments passed to the imaging generator.

        Returns
        -------
        dict or None
            Simulation output as a dictionary mapping modality names to results. Returns None if the imager could not be created.
        """
        np.random.seed(None)
        if not self.generators_status("imager"):
            self.build(modules="imager")
            if self.imager is None:
                print("Imager not created. Cannot run simulation.")
                return None
        if modality == "All":
            print("Simulating all modalities")
            if acq_params is None:
                acq_params = self.selected_mods
            if self.experiment_id:
                name = self.experiment_id
            simulation_output, simulation_output_noiseless = generate_multi_imaging_modalities(
                image_generator=self.imager,
                experiment_name=name,
                savingdir=self.output_directory,
                write=save,
                acquisition_param=acq_params,
            )
            self.results = simulation_output
            return simulation_output, simulation_output_noiseless
        else:
            print(f"Simulating: {modality}")
            acq_p = self.selected_mods[modality]
            timeseries_noise, beads_noise, timeseries_noiseless, beads_noiseless = self.imager.generate_imaging(
                modality=modality, **acq_p
            )
            simulation_output = {}
            simulation_output_noiseless = {}
            simulation_output[modality] = timeseries_noise
            simulation_output_noiseless[modality] = timeseries_noiseless
            return simulation_output, simulation_output_noiseless

    def remove_probes(self):
        """
        Remove all probe-related parameters and update the internal state accordingly.

        This method clears the probe parameters, updates the probes, and resets the
        particle and structure objects if their respective generators are active.
        It also clears any labels associated with the structure.

        Returns
        -------
        None
        """
        self.probe_parameters = dict()
        self._update_probes()
        if self.generators_status("particle"):
            self.particle = None
            self.objects_created["particle"] = False
            print("Labelled structure cleared")
        if self.generators_status("structure"):
            self.structure._clear_labels()
        print("Probes removed from experiment")

    def add_probe(
        self,
        probe_template: str = "NHS_ester",
        probe_name: str = None,
        probe_target_type: str = None,
        probe_target_value: str = None,
        probe_target_option: str = None,
        probe_distance_to_epitope: float = None,
        probe_model: str = None,
        probe_fluorophore: str = "AF647",
        probe_steric_hindrance=None,
        probe_paratope: str = None,
        probe_conjugation_target_info=None,
        probe_DoL: float = None,
        probe_seconday_epitope=None,
        probe_wobble_theta: float = None,
        labelling_efficiency: float = 1.0,
        as_primary=False,
        peptide_motif: dict = None,
        fluorophore_parameters = None,
        **kwargs,
    ):
        """
        Add a probe configuration to the experiment.

        This method loads a probe configuration from a YAML file, updates its parameters based on the provided arguments,
        and stores it in the experiment's probe parameters. It supports customization of probe properties such as target type,
        fluorophore, model, conjugation details, and more.

        Parameters
        ----------
        :param probe_name : str, optional
            Name of the probe. Default is "NHS_ester".
        :param probe_target_type : str, optional
            Type of the probe target (e.g., "Primary", "Secondary").
        :param probe_target_value : str, optional
            Value or identifier of the probe target.
        :param probe_target_option : str, optional
            Additional option for the probe target, used for secondary epitopes.
        :param probe_distance_to_epitope : float, optional
            Distance from the probe to the epitope.
        :param probe_model : str, optional
            List of model identifiers for the probe.
        :param probe_fluorophore : str, optional
            Identifier for the fluorophore. Default is "AF647".
        :param probe_steric_hindrance : Any, optional
            Steric hindrance value or configuration.
        :param probe_paratope : str, optional
            Paratope identifier or information.
        :param probe_conjugation_target_info : Any, optional
            Information about the conjugation target.
        :param probe_DoL : float, optional
            Efficiency of the probe conjugation.
        :param probe_seconday_epitope : Any, optional
            Information about a secondary epitope target.
        :param probe_wobbling : bool, optional
            Whether to enable probe wobbling. Default is False.
        :param labelling_efficiency : float, optional
            Efficiency of probe labelling. Default is 1.0.
        :param as_primary : bool, optional
            Whether to treat the probe as a primary linker. Default is False.
        :param **kwargs
            Additional keyword arguments for future extensions.

        Raises
        ------
        FileNotFoundError
            If the probe configuration YAML file does not exist.
        KeyError
            If required keys are missing in the configuration or probe parameters.

        Notes
        -----
        Updates the ``probe_parameters`` attribute with the new or modified probe configuration and calls the :meth:`_update_probes` method to refresh internal probe state.
        """
        probe_configuration = copy.deepcopy(
            self.config_probe_params[probe_template]
        )
        if probe_name is None:
            probe_name = probe_template
        else:
            probe_configuration["label_name"] = probe_name
        if peptide_motif is not None:
            protein_name, _1, site, sequence = (
                self.structure.get_peptide_motif(**peptide_motif)
            )
            if len(sequence) > 0:
                probe_target_type = "Sequence"
                probe_target_value = sequence
        if probe_target_type and probe_target_value:
            probe_configuration["target"] = dict(
                type=probe_target_type, value=probe_target_value
            )
            if probe_target_type == "Primary" and probe_target_option:
                if probe_target_value in self.probe_parameters.keys():
                    print(
                        f"Using {probe_target_option} as epitope on {probe_target_value}"
                    )
                    self.probe_parameters[probe_target_value][
                        "probe_seconday_epitope"
                    ] = probe_target_option
        elif (
            probe_configuration["target"]["type"] is None
            or probe_configuration["target"]["value"] is None
        ):
            print(
                "No target info provided for the probe. Retrieving random sequence."
            )
            # probe has no target info
            # a random target will be used
            protein_name, _1, site, sequence = (
                self.structure.get_peptide_motif(position="cterminal")
            )
            probe_configuration["target"] = dict(
                type="Sequence", value=sequence
            )
            # probe_configuration["target"]["type"] = "Sequence"
            # probe_configuration["target"]["value"] = sequence
        if probe_distance_to_epitope is not None:
            probe_configuration["distance_to_epitope"] = (
                probe_distance_to_epitope
            )
        if probe_fluorophore is not None:
            probe_configuration["fluorophore_id"] = probe_fluorophore
        if fluorophore_parameters is not None:
                self.add_fluorophore_parameters(fluorophoe_id=probe_fluorophore, **fluorophore_parameters)
        if labelling_efficiency is not None:
            probe_configuration["labelling_efficiency"] = labelling_efficiency
        if probe_model is not None:
            #probe_configuration["model_ID"] = probe_model
            probe_configuration["model"]["ID"] = probe_model
        if probe_paratope is not None:
            probe_configuration["paratope"] = probe_paratope
            probe_configuration["binding"]["paratope"] = probe_paratope
        if probe_conjugation_target_info is not None:
            probe_configuration["conjugation_target_info"] = (
                probe_conjugation_target_info
            )
            probe_configuration["conjugation_sites"]["target"]["type"] = (
                probe_conjugation_target_info["type"]
            )
            probe_configuration["conjugation_sites"]["target"]["value"] = (
                probe_conjugation_target_info["value"]
            )
        if probe_DoL is not None:
            print(f"Setting probe DoL to: {probe_DoL}")
            probe_configuration["conjugation_sites"]["DoL"] = (
                probe_DoL
            )
        else:
            print("Add_probe: Using default probe DoL from template")
        if probe_seconday_epitope is not None:
            probe_configuration["epitope_target_info"] = probe_seconday_epitope
        if probe_wobble_theta is not None:
            probe_configuration["enable_wobble"] = True
            probe_configuration["wobble_theta"] = probe_wobble_theta
        if as_primary:
            print("Adding probe as primary linker")
            probe_configuration["as_linker"] = True
        else:
            probe_configuration["as_linker"] = False
        if probe_steric_hindrance is not None:
            probe_configuration["distance_between_epitope"] = (
                probe_steric_hindrance
            )
            probe_configuration["binding"]["distance"][
                "between_targets"
            ] = probe_steric_hindrance
        self.probe_parameters[probe_name] = probe_configuration
        self._update_probes()

    def _update_probes(self):
        """
        Update the structure_label attribute based on the current probe_parameters.

        Returns
        -------
        None
        """
        if len(self.probe_parameters.keys()) == 0:
            self.structure_label = None
        else:
            self.structure_label = list(self.probe_parameters.keys())

    def add_fluorophore_parameters(self,
                                fluorophoe_id: str = None,
                                emission = None,
                                blinking_rates = None,
                                **kwargs):
        if fluorophoe_id not in self.fluorophore_parameters.keys():
            self.fluorophore_parameters[fluorophoe_id] = dict()
            self.fluorophore_parameters[fluorophoe_id]["emission"] = dict()
            self.fluorophore_parameters[fluorophoe_id]["emission"]["type"] = "constant"
            self.fluorophore_parameters[fluorophoe_id]["blinking_rates"] = dict()
            self.fluorophore_parameters[fluorophoe_id]["blinking_rates"]["kon"] = 0.99
            self.fluorophore_parameters[fluorophoe_id]["blinking_rates"]["koff"] = 0.01
            self.fluorophore_parameters[fluorophoe_id]["blinking_rates"]["kbleach"] = 0.0
            self.fluorophore_parameters[fluorophoe_id]["blinking_rates"]["initial_state"] = 1
            self.fluorophore_parameters[fluorophoe_id]["blinking_rates"]["photons_per_blink"] = 1000
        if emission is not None:
            self.fluorophore_parameters[fluorophoe_id]["emission"] = copy.deepcopy(emission)
        if blinking_rates is not None:
            self.fluorophore_parameters[fluorophoe_id]["blinking_rates"] = blinking_rates

    def set_virtualsample_params(
        self,
        virtualsample_template="square1x1um_randomised",
        sample_dimensions: list[int, int, int] = None,
        number_of_particles: int = None,
        particle_positions: list = None,
        sample_inital_orientation: list = None,
        random_orientations: bool = None,
        random_placing: bool = None,
        minimal_distance: float = None,
        update_mode: bool = True,
        random_rotations=None,
        rotation_angles=None,
        xy_orientations = None,
        xz_orientations = None,
        yz_orientations = None,
        axial_offset=None,
        rotation_per_particle=None,
        orientation_per_particle=None,
        **kwargs,
    ):
        """
        Set parameters for the virtual sample by loading a template configuration and optionally overriding specific values.

        Parameters
        ----------
        :param virtualsample_template : str, optional
            Name of the virtual sample template YAML file (without extension) to load. Default is "square1x1um_randomised".
        :param sample_dimensions : list of int, optional
            Dimensions of the sample to override the template value.
        :param number_of_particles : int, optional
            Number of particles to override the template value.
        :param particle_positions : list, optional
            List of particle positions to override the template value.
        :param random_orientations : bool, optional
            Whether to randomize particle orientations, overrides the template value.
        :param random_placing : bool, optional
            Whether to randomize particle placement, overrides the template value.
        :param **kwargs
            Additional keyword arguments (currently unused).

        Returns
        -------
        None
        """
        # load default configuration for virtual sample
        try:
            particle_minimal_distance = self.coordinate_field.molecules_params[
                "minimal_distance"
            ]
        except:
            particle_minimal_distance = None
        if update_mode:
            pass
        else:
            virtual_sample_template = os.path.join(
                self.configuration_path,
                "virtualsample",
                virtualsample_template + ".yaml",
            )
            self.virtualsample_params = load_yaml(virtual_sample_template)
        if sample_dimensions is not None:
            self.virtualsample_params["sample_dimensions"] = sample_dimensions
        if number_of_particles is not None:
            self.virtualsample_params["number_of_particles"] = number_of_particles
        if particle_positions is not None:
            self.virtualsample_params["relative_positions"] = particle_positions
        if random_orientations is not None:
            self.virtualsample_params["random_orientations"] = random_orientations
        if random_placing is not None:
            self.virtualsample_params["random_placing"] = random_placing
        if minimal_distance is not None:
            self.virtualsample_params["minimal_distance"] = minimal_distance
        else:
            self.virtualsample_params["minimal_distance"] = (
                particle_minimal_distance
            )
        if random_rotations is not None:
            self.virtualsample_params["random_rotations"] = random_rotations
        if rotation_angles is not None:
            self.virtualsample_params["rotation_angles"] = rotation_angles
        if xy_orientations is not None:
            self.virtualsample_params["xy_orientations"] = xy_orientations
        if xz_orientations is not None:
            self.virtualsample_params["xz_orientations"] = xz_orientations
        if yz_orientations is not None:
            self.virtualsample_params["yz_orientations"] = yz_orientations
        if axial_offset is not None:
            self.virtualsample_params["axial_offset"] = axial_offset
        if orientation_per_particle is not None:
            self.virtualsample_params["orientation_per_particle"] = orientation_per_particle
        if rotation_per_particle is not None:
            self.virtualsample_params["rotation_per_particle"] = rotation_per_particle
        if sample_inital_orientation is not None:
            if len(sample_inital_orientation) == 3:
                self.virtualsample_params["sample_inital_orientation"] = sample_inital_orientation
                self.virtualsample_params["random_rotations"] = False


    def use_image_for_positioning(
        self,
        img,
        mode="localmaxima",
        positions_are_epitopes=False,
        sigma=None,
        background=None,
        threshold=None,
        pixelsize=None,
        min_distance=None,
        elevation_img=None,
        normalise_elevation=True,
        max_elevation_nm=100,
        **kwargs,
    ):
        """
        Generate and set relative positions for a virtual sample based on features detected in an input image.

        This method processes the provided image to extract feature coordinates using the specified detection mode
        and parameters. The resulting positions are stored in `self.virtualsample_params["relative_positions"]`,
        and the sample's physical dimensions are updated accordingly. The method then triggers the build process
        for the 'coordinate_field' and 'imager' modules.

        Parameters
        ----------
        :param img : array-like
            The input image (as a NumPy array or compatible format) to be analyzed for feature detection.
        :param mode : str, optional
            The feature detection mode to use (e.g., "localmaxima"). Default is "localmaxima".
        :param sigma : float or None, optional
            Standard deviation for Gaussian smoothing. If None, no smoothing is applied.
        :param background : float or None, optional
            Background intensity to subtract from the image. If None, no subtraction.
        :param threshold : float or None, optional
            Minimum intensity threshold for feature detection. If None, uses default.
        :param pixelsize : float or None, optional
            Pixel size of the input image. If None, uses default.
        :param min_distance : float or None, optional
            Minimum distance between detected features. If None, uses default.
        :param **kwargs
            Additional keyword arguments passed to the feature detection function.

        Returns
        -------
        None

        Notes
        -----
        Updates `self.virtualsample_params` with new relative positions and sample dimensions. Calls `self.build()` for the specified modules.
        """
        xyz_relative, image_physical_size = (
            coordinates_field.gen_positions_from_image(
                img=img,
                mode=mode,
                sigma=sigma,
                background=background,
                threshold=threshold,
                pixelsize=pixelsize,
                min_distance=min_distance,
                elevation_img=elevation_img,
                normalise_elevation = normalise_elevation,
                **kwargs,
            )
        )
        if positions_are_epitopes:
            self.clear_structure()
            xyz_abs = np.array(xyz_relative)
            xyz_abs[:,0]*= image_physical_size[0]
            xyz_abs[:,1]*= image_physical_size[1]
            xyz_abs[:,2]*= max_elevation_nm
            print(xyz_abs.shape)
            normals = np.zeros(shape=xyz_abs.shape)
            normals[:,2] = 1
            probe_name = list(self.probe_parameters.keys())[0]
            particle_generator_data = dict(
                targets={
                    probe_name:dict(
                        coordinates = xyz_abs,
                        normals=normals
                    )
                },
                scale=1e-9,
                reference_point = np.array([image_physical_size[0]*0.5, image_physical_size[1]*0.5, 0]),
                axis=dict(pivot=np.array([0.0, 0.0, 0.0]), direction=np.array([0.0, 0.0, 1.0]))
            )
            self.particle.source = dict()
            # Input custom epitopes
            self.particle.load_source(**copy.deepcopy(particle_generator_data))
            self.particle.generate_instance()
            self.set_virtualsample_params(
                sample_dimensions=[
                    image_physical_size[0],
                    image_physical_size[1],
                    max_elevation_nm,
                ]
            )
            self.particle.generate_instance() # relabels particle based on current attributes
            # of build method is used it will insetead take info from available structure
            self.build(modules=["coordinate_field", "imager"])
        else:
            self.set_virtualsample_params(
                sample_dimensions=[
                    image_physical_size[0],
                    image_physical_size[1],
                    max_elevation_nm,
                ],
                particle_positions=xyz_relative,
                number_of_particles=len(xyz_relative),
                random_orientations=False,
                random_placing=False,
                minimal_distance=min_distance,
            )
            self.build(modules=["coordinate_field", "imager"])

    def current_settings(
        self, as_string=True, newline="<br>", modalities_acq_params=False
    ):
        """
        Print the current settings of the experiment, including structure, particle, coordinate field, and imaging modalities.

    Returns
    -------
    None
        """
        string = "Current settings of the experiment:" + newline
        string += f"Structure ID: {self.structure_id}" + newline
        string += f"Probes: {list(self.probe_parameters.keys())}" + newline
        string += f"Virtual sample: {self.coordinate_field_id}" + newline
        string += "Imaging Modalities: "
        for modality_name, acqparams in self.selected_mods.items():
            string += f"  {modality_name}"
            if modalities_acq_params:
                string += f": {acqparams}" + newline
            else:
                string += "; "
        string += newline
        if not as_string:
            print(string)
        else:
            return string

    def expand_virtual_sample(self, factor=1):
        if factor > 1:
            self.coordinate_field.expand_isotropically(factor=factor)
            self.build(modules=["imager",])

    def set_structure_axis_euler(self, phi = 0, psi=0, theta=0, order="zyx"):
        reference_vector = np.array([1,0,0]) # unitary vector in X
        reference_rotated = apply_euler_rotation(
            reference_vector, 
            phi = phi, 
            psi=psi, 
            theta=theta, 
            order="zyx")
        new_orientation_point = self.structure.axis["pivot"] + reference_rotated
        self.structure.set_axis_from_point(new_orientation_point)
    
    def set_structure_normal_params(self, mode = "scaling", normal_vector = None):
        self.structure.normals_params["mode"] = mode
        self.structure.normals_params["normal_vector"] = normal_vector


def generate_virtual_sample(
    structure: str = "1XI5",
    structure_is_path = False,
    structure_axis_euler:list = [0,0,0],
    structure_global_normal_orientation = None,
    probe_template: str = "NHS_ester",
    probe_name: str = None,
    probe_target_type: str = None,
    probe_target_value: str = None,
    probe_distance_to_epitope: float = None,
    probe_steric_hindrance: float = None,
    probe_model: list[str] = None,
    probe_fluorophore: str = "AF647",
    probe_paratope: str = None,
    probe_conjugation_target_info=None,
    probe_DoL: float = None,
    probe_seconday_epitope=None,
    probe_wobble_theta=None,
    labelling_efficiency: float = 1.0,
    structural_integrity_small_cluster: float = None,
    structural_integrity_large_cluster: float = None,
    structural_integrity: float = None,
    virtual_sample_template: str = "square1x1um_randomised",
    sample_dimensions: list[float] = None,
    sample_inital_orientation = None,
    number_of_particles: int = None,
    particle_positions: list[np.array] = None,
    orientation_per_particle = None,
    random_orientations = False,
    xy_orientations = None,
    xz_orientations = None,
    yz_orientations = None,
    axial_offset = None,
    random_placing=False,
    rotation_per_particle = None,
    random_rotations=False,
    rotation_angles = None,
    expansion_factor=1,
    clear_probes=False,
    clear_experiment=False,
    # secondary
    primary_probe=None,
    secondary_probe=None,
    probe_list=None,
    random_seed: int = None,
    **kwargs,
):
    """
    Create a virtual sample from a structure model.

    Parameters
    ----------
    :param structure : str, optional
        4-letter ID of PDB/CIF model. Default is "1XI5".
    :param structure_is_path : logical
        Use structure value as absolute path for the PDB/CIF file.
    :param structure_axis_euler: list
        Specify euler angles for structure central axis orientation with order zyx. For perspective angles the order is [Azimutal, Elevation, Roll]
    :param probe_name : str, optional
        Name ID of probe configuration file (filename).
    :param probe_target_type : str, optional
        Options: "Sequence", "Atom_residue", or "Primary".
    :param probe_target_value : str or dict, optional
        For target type "Sequence" or "Primary", a string. For "Atom_residue", a dictionary with keys "Atom" and "Residue".
    :param probe_distance_to_epitope : float, optional
        Minimal distance set from epitope and probe paratope.
    :param probe_model : list of str, optional
        4-letter ID(s) of PDB/CIF model(s).
    :param probe_fluorophore : str, optional
        Fluorophore name (e.g., "AF647"). Default is "AF647".
    :param probe_paratope : str, optional
        Sequence of the paratope site for when probe includes a model.
    :param probe_conjugation_target_info : Any, optional
        Information about the probe conjugation target.
    :param probe_DoL : float, optional
        Efficiency of conjugation of emitters.
    :param probe_seconday_epitope : str, optional
        Sequence within probe model to be used as epitope for a secondary.
    :param probe_wobbling : bool, optional
        Enable probe wobbling. Default is False.
    :param labelling_efficiency : float, optional
        Labelling efficiency of probe. Default is 1.0.
    :param structural_integrity_small_cluster : float, optional
        In Å, distance used to group epitopes into multimers.
    :param structural_integrity_large_cluster : float, optional
        In Å, distance within multimers to consider neighbors.
    :param structural_integrity : float, optional
        Fraction of structural_integrity to model.
    :param virtual_sample_template : str, optional
        Name of the configuration file for template. Default is "square1x1um_randomised".
    :param sample_dimensions : list of float, optional
        In nanometers, define the X, Y, and Z sizes of the field.
    :param number_of_particles : int, optional
        Number of independent copies of a particle to create and distribute.
    :param particle_positions : list of np.array, optional
        Relative positions of particles in the field.
    :param random_orientations : bool, optional
        If True, each particle will be randomly assigned a new orientation. Default is False.
    :param random_placing : bool, optional
        Define if position in field is random or the center of field. Default is False.
    :param clear_probes : bool, optional
        If True, default parameters will be cleared. Default is False.
    :param **kwargs
        Additional keyword arguments.

    Returns
    -------
    tuple
        - dict: The virtual sample as exported format. This can be used as input for the `image_vsample` method.
        - ExperimentParametrisation: The experiment containing all modules that were generated to build the virtual sample, and the virtual sample module itself. This experiment can be further used and tweaked for subsequent analysis or branching workflows.
    """
    myexperiment = ExperimentParametrisation(random_seed=random_seed)
    if clear_experiment:
        myexperiment.clear_experiment()
    # select structure
    if structure_is_path:
        print(f"Selecting structure from path: {structure}")
        myexperiment.select_structure(
            structure_id=structure.split(".")[0], 
            structure_path=structure,
            build=True
        )
    else:
        print("Selecting structure from ID:", structure)
        myexperiment.select_structure(
            structure_id=structure, 
            structure_path=None,
            build=True
        )
    if structure_axis_euler[0] or structure_axis_euler[1] or structure_axis_euler[2]:
        myexperiment.set_structure_axis_euler(*structure_axis_euler)
    if structure_global_normal_orientation is not None:
        if type(structure_global_normal_orientation) is str:
            myexperiment.set_structure_normal_params(
                mode=structure_global_normal_orientation)
        elif isinstance(structure_global_normal_orientation, np.ndarray):
            myexperiment.set_structure_normal_params(
                mode="global",
                normal_vector=structure_global_normal_orientation)
        else:
            print("set_structure_normal_params type error")
            print(isinstance(structure_global_normal_orientation, np.ndarray))
            
    # load default configuration for probe
    if (primary_probe is not None) and (secondary_probe is not None):
        print("Adding primary and secondary probes")
        myexperiment.add_probe(as_primary=True, **primary_probe)
        myexperiment.add_probe(as_primary=False, **secondary_probe)
    elif probe_list is not None:
        for probe in probe_list:
            print(f"Adding probe: {probe['probe_name']}")
            myexperiment.add_probe(**probe)
    else:
        if not clear_probes:
            print(probe_template)
            probe_configuration_file = os.path.join(
                myexperiment.configuration_path, "probes", probe_template + ".yaml"
            )
            probe_configuration = load_yaml(probe_configuration_file)
            probe_configuration["probe_template"] = probe_template
            if probe_name is None:
                probe_name = probe_template
            else:
                probe_configuration["label_name"] = probe_name
                probe_configuration["probe_name"] = probe_name
            if probe_target_type and probe_target_value:
                print(probe_target_type, probe_target_value)
                probe_configuration["probe_target_type"] = probe_target_type
                probe_configuration["probe_target_value"] = probe_target_value
            if probe_distance_to_epitope is not None:
                probe_configuration["probe_distance_to_epitope"] = (
                    probe_distance_to_epitope
                )
            if probe_steric_hindrance is not None:
                probe_configuration["probe_steric_hindrance"] = probe_steric_hindrance
            if probe_fluorophore is not None:
                probe_configuration["probe_fluorophore"] = probe_fluorophore
            if labelling_efficiency is not None:
                probe_configuration["labelling_efficiency"] = labelling_efficiency
            if probe_model is not None:
                probe_configuration["probe_model_ID"] = probe_model
            if probe_paratope is not None:
                probe_configuration["probe_paratope"] = probe_paratope
            if probe_conjugation_target_info is not None:
                probe_configuration["probe_conjugation_target_info"] = (
                    probe_conjugation_target_info
                )
            if probe_DoL is not None:
                probe_configuration["probe_DoL"] = (
                    probe_DoL
                )
            if probe_seconday_epitope is not None:
                probe_configuration["probe_epitope_target_info"] = probe_seconday_epitope
            if probe_wobble_theta is not None:
                probe_configuration["probe_wobble_theta"] = probe_wobble_theta
            myexperiment.add_probe(**probe_configuration)
    # load default configuration for virtual sample
    virtual_sample_template = os.path.join(
        myexperiment.configuration_path,
        "virtualsample",
        virtual_sample_template + ".yaml",
    )
    vsample_configuration = load_yaml(virtual_sample_template)
    #myexperiment.configuration_path
    if structural_integrity is not None and structural_integrity_large_cluster and structural_integrity_small_cluster:
        myexperiment.structural_integrity_eps["eps1"] = structural_integrity_small_cluster
        myexperiment.structural_integrity_eps["eps2"] = structural_integrity_large_cluster
        myexperiment.structural_integrity_eps["structural_integrity"] = structural_integrity
        myexperiment.structural_integrity_eps["use_structural_integrity"] = True

    if sample_dimensions is not None:
        vsample_configuration["sample_dimensions"] = sample_dimensions
    if sample_inital_orientation is not None:
        vsample_configuration["sample_inital_orientation"] = sample_inital_orientation
    else:
        vsample_configuration["sample_inital_orientation"] = None
    if number_of_particles is not None:
        vsample_configuration["number_of_particles"] = number_of_particles
    if particle_positions is not None:
        vsample_configuration["relative_positions"] = particle_positions
    if random_orientations is not None:
        vsample_configuration["random_orientations"] = random_orientations
    if random_placing is not None:
        vsample_configuration["random_placing"] = random_placing
    if random_rotations:
        vsample_configuration["random_rotations"] = random_rotations
    if orientation_per_particle is not None:
        vsample_configuration["orientation_per_particle"] = orientation_per_particle
    if rotation_per_particle is not None:
        vsample_configuration["rotation_per_particle"] = rotation_per_particle
    vsample_configuration["rotation_angles"] = rotation_angles
    vsample_configuration["xy_orientations"] = xy_orientations
    vsample_configuration["xz_orientations"] = xz_orientations
    vsample_configuration["yz_orientations"] = yz_orientations
    vsample_configuration["axial_offset"] = axial_offset
    myexperiment.virtualsample_params = vsample_configuration
    myexperiment.build(modules=[
                "particle",
                "coordinate_field",
                "imager"], use_locals=True)
    if expansion_factor > 1:
        myexperiment.expand_virtual_sample(factor=expansion_factor)
    # myexperiment.coordinate_field_id = virtual_sample
    return myexperiment.exported_coordinate_field, myexperiment


def build_virtual_microscope(
    modality="STED", multimodal: list[str] = None, experiment=None, random_seed=None, **kwargs
):
    """
    Initialise a virtual microscope for single or multimodal imaging.

    Parameters
    ----------
    :param modality : str, optional
        Modality name. Default is "STED".
    :param multimodal : list of str, optional
        List of modality names. Overrides the `modality` parameter if provided.
    :param experiment : ExperimentParametrisation, optional
        An Experiment object. If None, a new one is created.
    :param **kwargs
        Additional arguments passed to `add_modality`.

    Returns
    -------
    tuple
        - Imager: The virtual microscope with the specified modality models. Contains a default sample.
        - ExperimentParametrisation: The experiment containing the virtual microscope. All other modules are not initialised. This experiment can be further used and tweaked for subsequent analysis or branching workflows.
    """
    if experiment is None:
        experiment = ExperimentParametrisation(random_seed=random_seed)
    if multimodal is not None:
        for mod in multimodal:
            print(mod)
            experiment.add_modality(mod, **kwargs)
    else:
        experiment.add_modality(modality, **kwargs)
    if "use_local_field" in kwargs.keys():
        experiment._build_imager(use_local_field=kwargs["use_local_field"])
    else:
        experiment._build_imager()
    return experiment.imager, experiment


def image_vsample(
    vsample=None,
    modality: str = "STED",
    multimodal: list[str] = None,
    run_simulation: bool = True,
    structure: str = "1XI5",
    structure_is_path: bool = False,
    structure_axis_euler:list = [0,0,0],
    structure_global_normal_orientation = None,
    probe_template: str = "NHS_ester",
    probe_name: str = None,
    probe_target_type: str = None,
    probe_target_value: str = None,
    probe_distance_to_epitope: float = None,
    probe_steric_hindrance: float = None,
    probe_model: list = None,
    probe_fluorophore: str = "AF647",
    probe_paratope: str = None,
    probe_conjugation_target_info = None,
    probe_DoL: float = None,
    probe_seconday_epitope = None,
    probe_wobble_theta = None,
    labelling_efficiency: float = 1.0,
    structural_integrity_small_cluster: float = None,
    structural_integrity_large_cluster: float = None,
    structural_integrity: float = None,
    virtual_sample_template: str = "square1x1um_randomised",
    sample_dimensions: list = None,
    sample_inital_orientation = None,
    number_of_particles: int = None,
    particle_positions: list = None,
    orientation_per_particle = None,
    random_orientations = False,
    xy_orientations: list[int] = None,
    xz_orientations: list[int] = None,
    yz_orientations: list[int] = None,
    axial_offset: list[int] = None,
    random_placing = False,
    rotation_per_particle = None,
    random_rotations = False,
    rotation_angles = None,
    expansion_factor = 1,
    clear_probes = False,
    clear_experiment = False,
    primary_probe = None,
    secondary_probe = None,
    probe_list = None,
    save: bool = False,
    random_seed: int = None,
    **kwargs,
):
    """
    Generate imaging simulations of the specified virtual sample and imaging modalities.

    This function can either:
    1. Take an existing virtual sample (as a dictionary) and wrap it in a new experiment, adding the specified imaging modalities.
    2. Or, generate a new virtual sample and experiment from scratch using the provided parameters (matching those of `generate_virtual_sample`).

    Parameters
    ----------
    :param vsample : dict, optional
        Dictionary specifying sample parameters. Corresponds to Experiment attribute "exported_coordinate_field". If None, a new sample is generated.
    :param modality : str, optional
        Modality name to use for imaging. Default is "STED".
    :param multimodal : list of str, optional
        List of modality names. If provided, overrides the `modality` parameter and adds all listed modalities.
    :param run_simulation : bool, optional
        If True, generates image simulation for each modality set. Default is True.
    
    ### Parameters for virtual sample generation (see `generate_virtual_sample` for details):
    
    :param structure : str, optional
        4-letter ID of PDB/CIF model. Default is "1XI5".
    :param structure_is_path : bool, optional
        Use structure value as absolute path for the PDB/CIF file.
    :param structure_axis_euler: list
        Specify euler angles for structure central axis orientation with order zyx. For perspective angles the order is [Azimutal, Elevation, Roll]
    :param probe_template : str, optional
        Name of probe configuration file (filename). Default is "NHS_ester".
    :param probe_name : str, optional
        Name for the probe configuration.
    :param probe_target_type : str, optional
        Options: "Sequence", "Atom_residue", or "Primary".
    :param probe_target_value : str or dict, optional
        For target type "Sequence" or "Primary", a string. For "Atom_residue", a dictionary with keys "Atom" and "Residue".
    :param probe_distance_to_epitope : float, optional
        Minimal distance set from epitope and probe paratope.
    :param probe_model : list, optional
        4-letter ID(s) of PDB/CIF model(s).
    :param probe_fluorophore : str, optional
        Fluorophore name (e.g., "AF647"). Default is "AF647".
    :param probe_paratope : str, optional
        Sequence of the paratope site for when probe includes a model.
    :param probe_conjugation_target_info : any, optional
        Information about the probe conjugation target.
    :param probe_DoL : float, optional
        Efficiency of conjugation of emitters.
    :param probe_seconday_epitope : str, optional
        Sequence within probe model to be used as epitope for a secondary.
    :param probe_wobble_theta : any, optional
        Enable probe wobbling.
    :param labelling_efficiency : float, optional
        Labelling efficiency of probe. Default is 1.0.
    :param structural_integrity_small_cluster : float, optional
        In Å, distance used to group epitopes into multimers.
    :param structural_integrity_large_cluster : float, optional
        In Å, distance within multimers to consider neighbors.
    :param structural_integrity : float, optional
        Fraction of structural_integrity to model.
    :param virtual_sample_template : str, optional
        Name of the configuration file for template. Default is "square1x1um_randomised".
    :param sample_dimensions : list, optional
        In nanometers, define the X, Y, and Z sizes of the field.
    :param number_of_particles : int, optional
        Number of independent copies of a particle to create and distribute.
    :param particle_positions : list, optional
        Relative positions of particles in the field.
    :param random_orientations : bool, optional
        If True, each particle will be randomly assigned a new orientation. Default is False.
    :param xy_orientations, xz_orientations, yz_orientations : any, optional
        Orientation parameters for the sample.
    :param axial_offset : any, optional
        Axial offset for the sample.
    :param random_placing : bool, optional
        Define if position in field is random or the center of field. Default is False.
    :param random_rotations : bool, optional
        If True, apply random rotations to particles. Default is False.
    :param rotation_angles : any, optional
        Rotation angles for the sample.
    :param clear_probes : bool, optional
        If True, default probe parameters will be cleared. Default is False.
    :param clear_experiment : bool, optional
        If True, clear the experiment before generating a new sample. Default is False.
    :param primary_probe, secondary_probe : any, optional
        Dictionaries for primary and secondary probe configuration.

    Returns
    -------
    tuple
        - dict: Image simulations (per modality).
        - dict: Noiseless image simulations (per modality).
        - ExperimentParametrisation: The experiment object containing all modules and configuration.
    """
    if vsample is None:
        vsample, sample_experiment = generate_virtual_sample(
            structure=structure,
            structure_is_path=structure_is_path,
            structure_axis_euler=structure_axis_euler,
            structure_global_normal_orientation=structure_global_normal_orientation,
            probe_template=probe_template,
            probe_name=probe_name,
            probe_target_type=probe_target_type,
            probe_target_value=probe_target_value,
            probe_distance_to_epitope=probe_distance_to_epitope,
            probe_steric_hindrance=probe_steric_hindrance,
            probe_model=probe_model,
            probe_fluorophore=probe_fluorophore,
            probe_paratope=probe_paratope,
            probe_conjugation_target_info=probe_conjugation_target_info,
            probe_DoL=probe_DoL,
            probe_seconday_epitope=probe_seconday_epitope,
            probe_wobble_theta=probe_wobble_theta,
            labelling_efficiency=labelling_efficiency,
            structural_integrity_small_cluster=structural_integrity_small_cluster,
            structural_integrity_large_cluster=structural_integrity_large_cluster,
            structural_integrity=structural_integrity,
            virtual_sample_template=virtual_sample_template,
            sample_dimensions=sample_dimensions,
            sample_inital_orientation=sample_inital_orientation,
            number_of_particles=number_of_particles,
            particle_positions=particle_positions,
            orientation_per_particle=orientation_per_particle,
            random_orientations=random_orientations,
            xy_orientations=xy_orientations,
            xz_orientations=xz_orientations,
            yz_orientations=yz_orientations,
            axial_offset=axial_offset,
            random_placing=random_placing,
            rotation_per_particle=rotation_per_particle,
            random_rotations=random_rotations,
            rotation_angles=rotation_angles,
            clear_probes=clear_probes,
            clear_experiment=clear_experiment,
            primary_probe=primary_probe,
            secondary_probe=secondary_probe,
            probe_list=probe_list,
            expansion_factor=expansion_factor,
            random_seed=random_seed,
        )
        sample_experiment.clear_modalities()
        if multimodal is not None:
            for mod in multimodal:
                try:
                    mod_template = kwargs[mod]["modality_template"]
                    print(f"Adding modality: {mod}, {mod_template}")
                    sample_experiment.add_modality(mod, modality_template=mod_template)
                except:
                    sample_experiment.add_modality(mod)
        else:
            sample_experiment.add_modality(modality)
        sample_experiment.build(
            modules=[
                "imager",
            ]
        )
        for modality_name in sample_experiment.imaging_modalities.keys():
            if modality_name in kwargs.keys():
                sample_experiment.update_modality(
                    modality_name, **kwargs[modality_name]
                )
                sample_experiment.set_modality_acq(
                    modality_name, **kwargs[modality_name]
                )
    else:
        # need to create an experiment for it
        if multimodal is not None:
            vmicroscope, sample_experiment = build_virtual_microscope(
                multimodal=multimodal, use_local_field=True
            )
        else:
            vmicroscope, sample_experiment = build_virtual_microscope(
                modality=modality, use_local_field=True
            )
        for modality_name in sample_experiment.imaging_modalities.keys():
            if modality_name in kwargs.keys():
                sample_experiment.update_modality(
                    modality_name, **kwargs[modality_name]
                )
                sample_experiment.set_modality_acq(
                    modality_name, **kwargs[modality_name]
                )
        sample_experiment.imager.import_field(**vsample)
    imaging_output = dict()
    imaging_output_noiseless = dict()
    if run_simulation:
        imaging_output, imaging_output_noiseless = sample_experiment.run_simulation(save=save)
    return imaging_output, imaging_output_noiseless, sample_experiment
